# -*- coding: utf-8 -*-
"""Data storages, as the CSV file storage."""

import os
import pytz
import datetime
from .utils import detect_encoding, _sanitize_string, _is_list_of_integers, _to_float
from .units import TimeUnit
from .datastructures import TimePoint, DataTimePoint, DataTimeSlot, TimeSeries
from propertime.utils import dt_from_str, dt_from_s, s_from_dt, timezonize, now_dt
from .exceptions import NoDataException, FloatConversionError, ConsistencyException

# Setup logging
import logging
logger = logging.getLogger(__name__)

HARD_DEBUG = False

POSSIBLE_TIMESTAMP_labelS = ['timestamp', 'epoch']
NO_DATA_PLACEHOLDERS = ['na', 'nan', 'null', 'nd', 'undef']
DEFAULT_SLOT_DATA_LOSS = None


#======================
#  Base Storage class
#======================

class Storage(object):
    """The base storage class. Can be implemented to store one or more series. If storing more than one,
    then the id of which series to load or store must be provided in the ``get()`` and ``put()`` methods."""

    def get(self, id=None, start=None, end=None, *args, **kwargs):
        raise NotImplementedError()

    def put(self, series, id=None, *args, **kwargs):
        raise NotImplementedError()


#======================
#  CSV File Storage
#======================

class CSVFileStorage(Storage):
    """A CSV file storage. Supports both point and slot time series.

    The file encoding, the series type and the timestamp columns are all auto-detect with an heuristic approach
    by default, and asked to be set manually only if the heuristics fails. In particular, whether to create point or slot
    series is based on the sampling interval automatically detected: if this is above 24 hours, then slots are used.

    The header with the column labels is optional, and if not present the column numbers are used as labels.
    Comments in the CSV file are supported Comments in the CSV file only as full-line comments, where the line starts
    with one of the characters listed as comment character (``#`` and ``;`` by default).

    The ``timestamp_format``, ``date_format`` and ``time_format`` arguments can be set using Python strptime format codes
    (https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes), but the ``timestamp_format``
    also supports two special formats:

        * ``epoch``: the epoch timestamp, intended as the number of seconds from the 1st January 1970 UTC with decimals for sub-second precision.
        * ``iso8601``: the ISO 8601 timestamp format (see https://www.w3.org/TR/NOTE-datetime)

    Data can be simply comma-separated (or using a custom separator) as well as single or double quoted, check out the RFC4180
    (https://datatracker.ietf.org/doc/html/rfc4180) for more details on the CSV file format specification.

    Args:
        filename: the file name (including its path).
        timestamp_column: the column label to be used as timestamp. Either use this or the time_column and/or date_column parameters.
        timestamp_format: the timestamp column format.
        time_column: the column label to be used as the time part of the timestamp.
        time_format: the time column format.
        date_column: the column label to be used as the date part of the timestamp.
        date_format: the date column format.
        tz: the timezone on which to create the time series on. If the timestamps in the file are naive, then they are assumed on
            such timezone. If they are instead offset-aware, including UTC, then they are just moved on the given timezone.
        data_labels: the column labels to be used as data by default. Excpected as a list of
                     strings or list of integers, in which case are treated as column numbers.
        data_type: the data type (``list`` or ``dict``), set automatically by default.
        series_type: the default type of the series, if ``points`` or ``slots``. Set automatically by default.
        sort: if to sort the data before creating the series.
        separator: the separator for the records (fields), ``,`` by default.
        newline: the newline character, ``\\n`` by default.
        comment_chars: the characters used to mark a comment line. Defaulted to ``#`` and ``;``.
        encoding: the encoding of the file, set automatically by default.
        skip_errors: if to skip errors or raise an exception.
        silence_errors: if to completely silence errors when skipping them or not.
    """

    def __init__(self, filename, timestamp_column='auto', timestamp_format='auto',
                 time_column=None, time_format=None, date_column=None, date_format=None,
                 tz='UTC', data_labels='all', data_type='auto', series_type='auto', sort=False,
                 separator=',', newline='\n', comment_chars = ['#', ';'],  encoding='auto',
                 skip_errors=False, silence_errors=False):

        # Parameters sanity checks and adjustments
        if timestamp_column is None and time_column is None and date_column is None and not data_labels:
            raise ValueError('No timestamp column, time column, date column or data columns provided, cannot get anything from this CSV file')

        if timestamp_column != 'auto' and (time_column or date_column):
            raise ValueError('Got both timestamp column and time/date columns, choose which approach to use.')

        if timestamp_column is not None and not isinstance(timestamp_column, int) and not isinstance(timestamp_column, str):
            raise ValueError('timestamp_column argument must be a string (or integer for coulmn number) (got "{}")'.format(timestamp_column.__class__.__name__))

        if time_column is not None:
            if not isinstance(time_column, int) and not isinstance(time_column, str):
                raise ValueError('time_column argument must be a string (or integer for coulmn number) (got "{}")'.format(time_column.__class__.__name__))
            if not time_format:
                raise ValueError('If giving a time_column, a time_format is required as well')

            # Disble the timestamp label if a time (or date) label is set
            timestamp_column = None
            timestamp_format = None

        if date_column is not None:
            if not isinstance(date_column, int) and not isinstance(date_column, str):
                raise ValueError('date_column argument must be a string (or integer for coulmn number) (got "{}")'.format(date_column.__class__.__name__))
            if not date_format:
                raise ValueError('If giving a date_column, a date_format is required as well')

            # Disble the timestamp label if a date (or time) label is set
            timestamp_column = None
            timestamp_format = None

        if data_labels != 'all' and not isinstance(data_labels, list):
            raise ValueError('data_labels argument must be a list (got "{}")'.format(data_labels.__class__.__name__))

        # File & encoding
        self.filename = filename
        if encoding == 'auto':
            self.encoding = None
        else:
            self.encoding = encoding

        # Set time parameters
        self.timestamp_column = timestamp_column
        self.timestamp_format = timestamp_format

        self.date_column = date_column
        self.date_format = date_format

        self.time_column = time_column
        self.time_format = time_format

        # Timezone
        self.tz = tz

        # Data TODO: allow for columns mapping here? i.e data_labels = {1:'flow',2:'temperature'}
        self.data_labels = data_labels

        # Separator, newline and comment chars
        self.separator = separator
        self.newline = newline
        self.comment_chars = comment_chars

        # Other
        self.skip_errors = skip_errors
        self.silence_errors = silence_errors
        self.data_type = data_type
        self.sort = sort

        # Set series type (that will be forced)
        self.series_type = series_type


    def get(self, id=None, start=None, end=None, limit=None, filter_data_labels=[],
            force_tz=None, force_points=False, force_slots=False, force_slot_unit=None):
        """Load the time series from the CSV file.

        Args:
            id: Not implemented for this storage.
            start: Not implemented for this storage.
            end: Not implemented for this storage.
            limit: a row number limit.
            filter_data_labels: get only specific data labels.
            force_tz: force a specific timezone.
            force_points: force generating points.
            force_slots: force generating slots.
            force_slot_unit: set the unit of the slots.
        """

        if id:
            raise NotImplementedError('This storage does not support multiple time series, so the id argument cannot be used')

        if start or end:
            raise NotImplementedError('This storage does not support loading only a portion of the time time series, so the start and end arguments cannot be used')

        # Sanity checks
        if force_points and force_slots:
            raise ValueError('Got both as_points and as_slots, set only one or none')

        # Line counter
        line_number=0

        # Init column indexes and labels
        column_indexes = None
        column_labels = None

        timestamp_column_index = None
        time_column_index = None
        date_column_index = None

        data_label_indexes = None
        data_label_names = None

        # Data type
        data_type = None

        # Detect encoding if not set
        if not self.encoding:
            self.encoding = detect_encoding(self.filename, streaming=False)

        # Set and timezonize the timezone. Using this in the points or slots will be just a pointer
        if force_tz:
            tz = timezonize(force_tz)
        else:
            tz = timezonize(self.tz)

        # Loop over all CSV rows
        # TODO: evaluate rU vs newline='\n'
        items = []
        with open(self.filename, 'r', encoding=self.encoding) as csv_file:

            while True:

                # TODO: replace me using a custom line separator
                line = csv_file.readline()
                logger.debug('Processing line #%s: "%s"', line_number, _sanitize_string(line,NO_DATA_PLACEHOLDERS))

                # Do we have to stop? Note: empty lines still have the "\n" char.
                if not line:
                    break

                # Is this line an empty line?
                #if not line.strip():
                #    continue

                # Is this line a comment?
                if line[0] in self.comment_chars:
                    logger.debug('Skipping line "%s" as marked as comment line', line)
                    continue

                # Is this line something else?
                if not self.separator in line:
                    logger.debug('Skipping line "%s" as marked no value separator found', line)
                    continue

                # Get values from this line:
                line_items = line.split(self.separator)

                # Set column keys if not already done
                if not column_indexes:

                    # Is this line carrying non-numerical or non timestamp data?
                    not_converted = 0
                    for value in line_items:
                        try:
                            float(_sanitize_string(value,NO_DATA_PLACEHOLDERS))
                        except:
                            try:
                                dt_from_str(_sanitize_string(value,NO_DATA_PLACEHOLDERS))
                            except:
                                not_converted += 1

                    # If it is, use it as labels (and continue with the next line)
                    if not_converted: # == len(line_items):
                        column_indexes = [i for i in range(len(line_items)) if _sanitize_string(line_items[i],NO_DATA_PLACEHOLDERS)]
                        column_labels = [_sanitize_string(label,NO_DATA_PLACEHOLDERS) for label in line_items if _sanitize_string(label,NO_DATA_PLACEHOLDERS)]
                        logger.debug('Set column indexes = "%s"  and column labels = "%s"', column_indexes, column_labels)
                        continue

                    # Otherwise, just use integer keys (and go on)
                    else:
                        column_indexes = [i for i in range(len(line_items)) if _sanitize_string(line_items[i],NO_DATA_PLACEHOLDERS)]
                        logger.debug('Set column indexes = "%s"  and column labels = "%s"', column_indexes, column_labels)

                    # Note: above we only used labels and indexes that carry content after being sanitize_stringd.
                    # TODO: what if a file starts with a line missing last value and we do or not do set data columns from the outside?
                    #       .. external data columns should have precedenze, but a mapping is then required more than just a list.

                #====================
                #  Time part
                #====================

                # Set timestamp column index if not already done
                if self.timestamp_column is not None and timestamp_column_index is None:
                    if self.timestamp_column=='auto':
                        # TODO: improve this auto-detect: try different column names and formats.
                        #       Maybe, fix position as the first element for the timestamp.
                        if column_labels:
                            for posssible_timestamp_column_name in POSSIBLE_TIMESTAMP_labelS:
                                if posssible_timestamp_column_name in column_labels:
                                    timestamp_column_index = column_labels.index(posssible_timestamp_column_name)
                                    self.timestamp_column = posssible_timestamp_column_name
                                    break
                            if timestamp_column_index is None:
                                #raise Exception('Cannot auto-detect timestamp column')
                                timestamp_column_index = 0
                        else:
                            timestamp_column_index = 0
                        # TODO: Try to convert to timestamp here?
                    else:
                        if isinstance(self.timestamp_column, int):
                            timestamp_column_index = self.timestamp_column
                        else:
                            if self.timestamp_column in column_labels:
                                timestamp_column_index = column_labels.index(self.timestamp_column)
                            else:
                                if self.timestamp_column is not None:
                                    raise Exception('Cannot find requested timestamp column "{}" in labels (got "{}")'.format(self.timestamp_column, column_labels))
                                elif self.timestamp_column is not None:
                                    raise Exception('Cannot find requested time column "{}" in labels (got "{}")'.format(self.timestamp_column, column_labels))
                                else:
                                    pass
                    logger.debug('Set time column index = "%s"', timestamp_column_index)

                # Set time column index if required and not already done
                if self.time_column is not None and time_column_index is None:
                    if isinstance(self.time_column, int):
                        time_column_index = self.time_column
                    else:
                        if self.time_column in column_labels:
                            time_column_index = column_labels.index(self.time_column)
                        else:
                            raise Exception('Cannot find requested time column "{}" in labels (got "{}")'.format(self.time_column, column_labels))
                    logger.debug('Set date column index = "%s"', time_column_index)

                # Set date column index if required and not already done
                if self.date_column is not None and date_column_index is None:
                    if isinstance(self.date_column, int):
                        date_column_index = self.date_column
                    else:
                        if self.date_column in column_labels:
                            date_column_index = column_labels.index(self.date_column)
                        else:
                            raise Exception('Cannot find requested time column "{}" in labels (got "{}")'.format(self.time_column, column_labels))
                    logger.debug('Set date column index = "%s"', date_column_index)

                # If all the three of timestamp_column_index, time_column_index and date_column_index are None, there is something wrong:
                if timestamp_column_index is None and time_column_index is None and date_column_index is None:
                    raise ConsistencyException('Could not set timestamp_column_index nor time_column_index nor date_column_index, somehting wrong happened.')

                # Ok,now get the timestamp as string
                if timestamp_column_index is not None:
                    # Just use the Timestamp column
                    timestamp = _sanitize_string(line_items[timestamp_column_index],NO_DATA_PLACEHOLDERS)

                else:
                    # Time part
                    time_part=None
                    if time_column_index is not None:
                        time_part = _sanitize_string(line_items[time_column_index],NO_DATA_PLACEHOLDERS)

                    # Date part
                    date_part=None
                    if date_column_index is not None:
                        date_part = _sanitize_string(line_items[date_column_index],NO_DATA_PLACEHOLDERS)

                    # Assemble timestamp
                    if time_part is not None and date_part is not None:
                        timestamp = date_part + '\t' + time_part
                    elif date_part is None:
                        timestamp = time_part
                    elif time_part is None:
                        timestamp = date_part
                    else:
                        raise ConsistencyException('Got both time_part and date_part as None, somehting wrong happened.')

                # Convert timestamp to epoch seconds
                logger.debug('Will process timestamp "%s"', timestamp)
                try:

                    t = None
                    dt = None

                    # Both time and date labels
                    if self.time_column is not None and self.date_column is not None:
                        dt = datetime.datetime.strptime(timestamp, self.date_format + '\t' + self.time_format)

                    # Only date label
                    elif self.date_column is not None:
                        dt = datetime.datetime.strptime(timestamp, self.date_format)

                    # Only time label (TODO: does this make sense?)
                    elif self.time_column is not None:
                        dt = datetime.datetime.strptime(timestamp, self.time_format)

                    # Use the timestamp label and format
                    else:

                        # Autodetect?
                        if self.timestamp_format == 'auto':
                            logger.debug('Auto-detecting timestamp format')

                            # Is this an epoch?
                            try:
                                t = float(timestamp)
                                logger.debug('Auto-detected timestamp format: epoch')
                                self.timestamp_format = 'epoch'
                            except:
                                try:
                                    # is this an iso8601?
                                    dt = dt_from_str(timestamp)
                                    logger.debug('Auto-detected timestamp format: iso8601')
                                    self.timestamp_format = 'iso8601'
                                except:
                                    raise Exception('Cannot auto-detect timestamp format (got "{}")'.format(timestamp)) from None

                        # Epoch format?
                        elif self.timestamp_format == 'epoch':
                            t = float(timestamp)

                        # ISO8601 format?
                        elif self.timestamp_format == 'iso8601':
                            dt = dt_from_str(timestamp)

                        # Custom format?
                        else:
                            dt = datetime.datetime.strptime(timestamp, self.timestamp_format)


                    # Convert to t
                    if t is None:
                        if dt.tzinfo is None:
                            dt = tz.localize(dt)
                        t = s_from_dt(dt)


                except Exception as e:
                    if self.skip_errors:
                        logger.debug('Cannot parse timestamp "%s" (%s) in line #%s, skipping the line.', timestamp, e, line_number)
                        continue
                    else:
                        raise ValueError('Cannot parse timestamp "{}" ({}) in line #%{}, aborting. Set "skip_errors=True" to drop them instead.'.format(timestamp, e, line_number)) from None

                #====================
                #  Data part
                #====================

                if data_label_indexes is None:

                    # If the requested data labels are all o the (default) ones, use the internal value
                    if not filter_data_labels:
                        data_labels = self.data_labels
                    else:
                        data_labels = filter_data_labels

                    # Do we have to select only some data columns?
                    if data_labels != 'all':

                        if _is_list_of_integers(data_labels):
                            # Case where we have been given a list of integers
                            data_label_indexes = data_labels
                            self.data_type = list

                        else:
                            # Case where we have been given a list of labels
                            if not column_labels:
                                raise Exception('You asked for data column labels but there are no labels in this CSV file')
                            data_label_indexes = []
                            for data_label in data_labels:
                                if data_label not in column_labels:
                                    raise Exception('Cannot find data column "{}" in CSV columns "{}"'.format(data_label, column_labels))
                                # What is the position of this requested data column in the CSV columns?
                                data_label_indexes.append(column_labels.index(data_label))

                    else:
                        # Case where we have to automatically set data column indexes to include all data columns

                        # Initialize the data_label_indexes equal to the column_indexes
                        data_label_indexes = column_indexes

                        # Remove timestamp, time and date indexes from the data_label_indexes
                        if timestamp_column_index is not None:
                            data_label_indexes.remove(timestamp_column_index)

                        if time_column_index is not None:
                            data_label_indexes.remove(time_column_index)

                        if date_column_index is not None:
                            data_label_indexes.remove(date_column_index)

                    # Ok, now based on the data_label_indexes fill the data_label_names if we have column labels
                    if column_labels:
                        data_label_names=[]
                        for data_label_index in data_label_indexes:
                            data_label_names.append(column_labels[data_label_index])

                    logger.debug('Set data_label_indexes="%s" and data_label_names="%s"', data_label_indexes, data_label_names)

                # Set data type
                if not data_type:
                    if data_label_names and len(data_label_names) > 1 and self.data_type == float:
                        raise Exception('Requested data format as float but got more than 1 value')
                    if self.data_type=='auto':
                        if not data_label_names:
                            data_type = list
                        else:
                            data_type = dict
                    else:
                        data_type = self.data_type

                # Set data
                try:

                    if data_type == float:
                        data = _to_float(line_items[data_label_indexes[0]],NO_DATA_PLACEHOLDERS)
                    elif data_type == list:
                        data = [_to_float(line_items[index],NO_DATA_PLACEHOLDERS) for index in data_label_indexes]
                    elif data_type == dict:
                        data = {column_labels[index]: _to_float(line_items[index],NO_DATA_PLACEHOLDERS,column_labels[index]) for index in data_label_indexes}

                # TODO: here we drop the entire line. Instead, should we use a "None" and allow "Nones" in the DataPoints data?
                except FloatConversionError as e:
                    if self.skip_errors:
                        logger.debug('Cannot convert value "%s" in line #%s to float, skipping the line.', e, line_number)
                        continue
                    else:
                        raise Exception('Cannot convert value "{}" in line #{} to float, aborting. Set "skip_errors=True" to drop them instead.'.format(e, line_number)) from None

                except IndexError as e:
                    if self.skip_errors:
                        logger.debug('Cannot parse in line #%s as some values are missing, skipping the line.', line_number)
                        continue
                    else:
                        raise Exception('Cannot parse line #{} as some values are missing, aborting. Set "skip_errors=True" to drop them instead.'.format(line_number)) from None

                logger.debug('Set data to "%s"', data)

                # Append to the series
                items.append([t, data])

                # Lastly, increase line_counter and check for limit
                line_number+=1
                if limit and line_number >= limit:
                    break

        # Were we able to read something?
        if not items:
            raise NoDataException('Cannot read any data!')

        # Sort if required
        if self.sort:
            from operator import itemgetter
            items = sorted(items, key=itemgetter(0))

        # Support var
        autodetect_series_type = False

        # Set series type
        if force_points:
            # Were we requested to generate points?
            series_type = 'points'
        elif force_slots:
            # Were we requested to generate slots?
            series_type = 'slots'
        else:
            # Do we have a series type set?
            if self.series_type != 'auto':
                # Use the object one
                series_type = self.series_type
            else:
                # Go in auto mode
                series_type = None

        # Otherwise, auto-detect
        if not series_type or series_type=='slots':

            autodetect_series_type = True

            # Detect sampling interval to create right item type (points or slots)
            from .utils import detect_sampling_interval

            # Add first ten elements max and then detect sampling interval.
            # Use a try-execpt as this can lead to errors due to duplicates,
            sample_timepointseries = TimeSeries()
            for item in items:
                try:
                    sample_timepointseries.append(TimePoint(t=item[0]))
                except:
                    pass
                if len(sample_timepointseries)>10:
                    break
            detected_sampling_interval = detect_sampling_interval(sample_timepointseries)

            # Years
            if detected_sampling_interval in [86400*365, 886400*366]:
                detected_series_type = DataTimeSlot
                auto_slot_unit = TimeUnit('1Y')

            # Months
            elif detected_sampling_interval in [86400*31, 86400*30, 86400*28]:
                detected_series_type = DataTimeSlot
                auto_slot_unit = TimeUnit('1M')

            # Days
            elif detected_sampling_interval in [3600*24, 3600*23, 3600*25]:
                detected_series_type = DataTimeSlot
                auto_slot_unit = TimeUnit('1D')

            # Weeks still to be implemented in the unit
            #elif detected_sampling_interval in [3600*24*7, (3600*24*7)-3600, (3600*24*7)+3600]:
            #    detected_series_type = DataTimeSlot
            #    auto_slot_unit = TimeUnit('1D')

            # Else, use points with no unit if we were not using slots
            else:
                if series_type!='slots':
                    detected_series_type = DataTimePoint
                    auto_slot_unit = None
                else:
                    # TODO: "detected_series_type" is not a nice name here, the
                    # code in the following should use series_type if forced..
                    detected_series_type = DataTimeSlot

                    # Can we auto-detect unit? TODO: can we standardize this? Check also in the entire codebase..
                    if detected_sampling_interval == 3600:
                        auto_slot_unit = TimeUnit('1h')
                    elif detected_sampling_interval == 1800:
                        auto_slot_unit = TimeUnit('30m')
                    elif detected_sampling_interval == 900:
                        auto_slot_unit = TimeUnit('15m')
                    elif detected_sampling_interval == 600:
                        auto_slot_unit = TimeUnit('10m')
                    elif detected_sampling_interval == 300:
                        auto_slot_unit = TimeUnit('5m')
                    elif detected_sampling_interval == 60:
                        auto_slot_unit = TimeUnit('1m')
                    else:
                        auto_slot_unit = TimeUnit('{}s'.format(detected_sampling_interval))

        # Do we have to force a specific type?
        if not autodetect_series_type:
            if series_type == 'points':
                series_type = DataTimePoint
                slot_unit = None
            elif series_type == 'slots':
                series_type = DataTimeSlot
                slot_unit = TimeUnit('{}s'.format(detected_sampling_interval))
            else:
                raise ValueError('Unknown value "{}" for type. Accepted types are "points" or "slots".'.format(self.series_type))
        else:
            # Log the type and unit we detected
            if detected_series_type == DataTimeSlot:
                if series_type:
                    logger.info('Assuming {} time unit and creating Slots.'.format(auto_slot_unit))
                else:
                    logger.info('Assuming {} time unit and creating Slots. Use series_type=\'points\' if you want Points instead.'.format(auto_slot_unit))
            #else:
            #    logger.info('Assuming {} sampling interval and creating {}.'.format(detected_sampling_interval, series_type.__class__.__name__))

            # and use it.
            series_type = detected_series_type
            slot_unit = auto_slot_unit

        # If we were explicitly given a slot unit, override
        if force_slot_unit:
            if not isinstance(force_slot_unit, TimeUnit):
                force_slot_unit = TimeUnit(force_slot_unit)
            slot_unit = force_slot_unit

        # Initialize the series
        series = TimeSeries()

        # Create point or slot series
        if series_type == DataTimePoint:

            for item in items:
                try:
                    # Handle data_indexes (including the data_loss)
                    data_indexes = None
                    if isinstance (item[1], dict):
                        data_indexes = {}
                        for key in item[1]:
                            if key.startswith('__'):
                                data_indexes[key[2:]] = item[1][key]
                        # Remove data_indexes from item data
                        for index in data_indexes:
                            item[1].pop('__'+index)

                    # Create DataTimePoint, set data and data_indexes
                    datatimepoint = DataTimePoint(t=item[0], data=item[1], data_indexes=data_indexes, tz=tz)

                    # Append
                    series.append(datatimepoint)

                except Exception as e:
                    if self.skip_errors:
                        if not self.silence_errors:
                            logger.error(e)
                    else:
                        raise e from None
        else:

            for i, item in enumerate(items):
                try:
                    # Handle data_indexes (including the data_loss)
                    data_indexes = None
                    if isinstance (item[1], dict):
                        data_indexes = {}
                        for key in item[1]:
                            if key.startswith('__'):
                                data_indexes[key[2:]] = item[1][key]
                        # Remove data_indexes from item data
                        for index in data_indexes:
                            item[1].pop('__'+index)

                    if DEFAULT_SLOT_DATA_LOSS is not None and 'data_loss' not in data_indexes:
                        data_indexes['data_loss'] = DEFAULT_SLOT_DATA_LOSS

                    # Create DataTimeSlot, set data and data_indexes
                    datatimeslot = DataTimeSlot(t=item[0], unit=slot_unit, data=item[1], data_indexes=data_indexes, tz=tz)

                    # Append
                    series.append(datatimeslot)

                except ValueError as e:
                    # The only ValueError that could (should) arise here is a "Not in succession" error.
                    missing_timestamps = []
                    prev_dt = dt_from_s(items[i-1][0], tz=tz)
                    while True:
                        dt = prev_dt + slot_unit
                        # Note: the equal here is just to prevent endless loops, the check shoudl actally be just an equal
                        if s_from_dt(dt) >= item[0]:
                            # We are arrived, append all the missing items and then the item we originally tried to and break
                            for j, missing_timestamp in enumerate(missing_timestamps):
                                # Set data by interpolation
                                interpolated_data = {data_label: (((items[i][1][data_label]-items[i-1][1][data_label])/(len(missing_timestamps)+1)) * (j+1)) + items[i-1][1][data_label]  for data_label in  items[-1][1] }
                                series.append(DataTimeSlot(t=missing_timestamp, unit=slot_unit, data=interpolated_data, data_loss=1, tz=tz))
                            series.append(DataTimeSlot(t=item[0], unit=slot_unit, data=item[1], data_loss=DEFAULT_SLOT_DATA_LOSS, tz=tz))
                            break

                        else:
                            missing_timestamps.append(s_from_dt(dt))
                        prev_dt = dt

        return series


    def put(self, series, id=None, overwrite=False):
        """Store the time series in the CSV file.

        Args:
            id: Not implemented for this storage.
            overwrite: if the destination file can be overwritten.
        """
        if id:
            raise NotImplementedError('This storage does not support multiple time series, so the id argument cannot be used')

        if os.path.isfile(self.filename) and not overwrite:
            raise Exception('File already exists. use overwrite=True to overwrite.')

        # Set data_indexes here once for all or there will be a slowdown afterwards
        data_indexes = series._all_data_indexes()

        with open(self.filename, 'w') as csv_file:

            # 0) Dump CSV-storage specific metadata & chck type
            if issubclass(series.item_type, DataTimePoint):
                csv_file.write('# Generated by Timeseria CSVFileStorage at {}\n'.format(now_dt()))
                csv_file.write('# {}\n'.format(series))
                csv_file.write('# Extra parameters: {{"type": "points", "tz": "{}", "resolution": "{}"}}\n'.format(series.tz, series.resolution))

            elif issubclass(series.item_type, DataTimeSlot):
                csv_file.write('# Generated by Timeseria CSVFileStorage at {}\n'.format(now_dt()))
                csv_file.write('# {}\n'.format(series))
                csv_file.write('# Extra parameters: {{"type": "slots", "tz": "{}", "resolution": "{}"}}\n'.format(series.tz, series.resolution))

            else:
                raise TypeError('Can store only time series of DataTimePoints or DataTimeSlots')

            # 1) Dump headers
            data_labels_part = ','.join([str(data_label) for data_label in series.data_labels()])
            data_indexes_part = ','.join(['__'+index for index in data_indexes])
            if data_indexes_part:
                csv_file.write('epoch,{},{}\n'.format(data_labels_part,data_indexes_part))
            else:
                csv_file.write('epoch,{}\n'.format(data_labels_part))

            # 2) Dump data (and data_indexes)
            for item in series:
                if isinstance (series[0].data, list) or isinstance (series[0].data, tuple):
                    data_part = ','.join([str(item.data[int(data_label)]) for data_label in series.data_labels()])
                else:
                    data_part = ','.join([str(item.data[data_label]) for data_label in series.data_labels()])
                data_indexes_part = ','.join([str(item.data_indexes[index]) for index in data_indexes])
                if data_indexes_part:
                    csv_file.write('{},{},{}\n'.format(item.t, data_part, data_indexes_part))
                else:
                    csv_file.write('{},{}\n'.format(item.t, data_part))







# -*- coding: utf-8 -*-
"""Series transformations as resampling and aggregation."""

from propertime.utils import dt_from_s, s_from_dt, as_tz
from datetime import datetime
from .datastructures import Point, Slot, DataTimeSlot, TimePoint, DataTimePoint, TimeSeries, TimeSeriesView
from .utils import _compute_data_loss, _compute_validity_regions
from .operations import avg
from .units import TimeUnit
from .exceptions import ConsistencyException
from . import operations as operations_module
from .interpolators import LinearInterpolator

# Setup logging
import logging
logger = logging.getLogger(__name__)


#==============================
# Support classes and functions
#==============================

class _TimeSeriesDenseView(TimeSeriesView):
    """A time series view which is "dense". Only used internally.

    :meta private:
    """
    def __init__(self, series, from_i, to_i, from_t=None, to_t=None, dense=False, interpolator_class=None):
        self.series = series
        self.from_i = from_i
        self.to_i = to_i
        self.len = None
        self.new_points = {}
        self.from_t = from_t
        self.to_t=to_t
        self.dense=dense
        if self.dense:
            if not interpolator_class:
                raise ValueError('If requesting a dense view you must provide an interpolator')
            self.interpolator = interpolator_class(series)
        else:
            self.interpolator = None

    def __getitem__(self, i):
        if self.dense:
            raise NotImplementedError('Getting items by index on dense views is not supported. Use the iterator instead.')

        if i>=0:
            return self.series[self.from_i + i]
        else:
            return self.series[self.to_i - abs(i)]

    def __iter__(self):
        self.count = 0
        self.prev_was_new = False
        return self

    def __next__(self):

        this_i = self.count + self.from_i

        if this_i >= self.to_i:
            # If reached the end stop
            raise StopIteration

        elif self.count == 0 or not self.dense:
            # If first point or not dense just return
            self.count += 1
            return self.series[this_i]

        else:
            # Otherwise check if we have to add new missing points

            if self.prev_was_new:
                # If we just created a new missing point, return
                self.prev_was_new = False
                self.count += 1
                return self.series[this_i]

            else:
                # Check if we have to add a new missing point: do we have a gap?
                prev_point = self.series[this_i-1]
                this_point = self.series[this_i]


                if prev_point.valid_to < this_point.valid_from:

                    # yes, we do have a gap. Add a missing point by interpolation

                    # Compute new point validity
                    if self.from_t is not None and prev_point.valid_to < self.from_t:
                        new_point_valid_from = self.from_t
                    else:
                        new_point_valid_from = prev_point.valid_to

                    if self.to_t is not None and this_point.valid_from > self.to_t:
                        new_point_valid_to = self.to_t
                    else:
                        new_point_valid_to = this_point.valid_from

                    # Compute new point timestamp
                    new_point_t = new_point_valid_from + (new_point_valid_to-new_point_valid_from)/2

                    # Can we use cache?
                    if new_point_t in self.new_points:
                        self.prev_was_new = True
                        return self.new_points[new_point_t]

                    # Log new point creation
                    logger.debug('New point t=,%s validity: [%s,%s]',new_point_t, new_point_valid_from,new_point_valid_to)

                    # Compute the new point values using the interpolator
                    new_point_data = self.interpolator.evaluate(new_point_t, prev_i=this_i-1, next_i=this_i)

                    # Create the new point
                    new_point = this_point.__class__(t = new_point_t, data = new_point_data)
                    new_point.valid_from = new_point_valid_from
                    new_point.valid_to = new_point_valid_to
                    new_point._interpolated = True

                    # Set flag
                    self.prev_was_new = True

                    # Add to cache
                    self.new_points[new_point_t] = new_point

                    # ..and return it
                    return new_point

                else:
                    # Return this point if no gaps
                    self.count += 1
                    return this_point

    def __len__(self):
        if not self.dense:
            return self.to_i-self.from_i
        else:
            if self.len is None:
                self.len=0
                for _ in self:
                    self.len+=1
            return self.len

    def __repr__(self):
        if not self.series:
            return 'Empty time series view'
        else:
            return 'Time series view'

    @property
    def item_type(self):
        for item in self:
            return item.__class__

    @property
    def resolution(self):
        return self.series.resolution

    def data_labels(self):
        return self.series.data_labels()


def _compute_new(target, series, from_t, to_t, slot_first_point_i, slot_last_point_i, slot_prev_point_i, slot_next_point_i,
                 unit, point_validity, timezone, fill_with, force_data_loss, fill_gaps, series_data_indexes, series_resolution,
                 force_compute_data_loss, interpolator_class, operations=None):
    """Support function for computing new items.

    :meta private:
    """

    # Log. Note: if slot_first_point_i < slot_last_point_i, this means that the prev and next are outside the slot.
    # It is not a bug, it is how the system works. perhaps we could pass here the slot_prev_i and sòlot_next_i
    # to make it more clear to avoid some confusion.
    logger.debug('Called compute new')
    logger.debug('slot_first_point_i=%s, slot_last_point_i=%s, ', slot_first_point_i, slot_last_point_i)
    logger.debug('slot_prev_point_i=%s, slot_next_point_i=%s, ', slot_prev_point_i, slot_next_point_i)

    # Support vars
    interval_duration = to_t-from_t
    data = {}
    data_labels = series.data_labels()

    # The prev can be None as lefts are included (edge case)
    if slot_prev_point_i is None:
        slot_prev_point_i = slot_first_point_i

    # Create a view of the series containing the slot datapoints plus the prev and next,
    series_dense_view_extended  = _TimeSeriesDenseView(series, from_i=slot_prev_point_i, to_i=slot_next_point_i+1,  # Slicing exclude the right
                                                       from_t=from_t, to_t=to_t, dense=True, interpolator_class=interpolator_class)

    # Compute the data loss for the new element. This is forced
    # by the resampler or slotter if first or last point
    data_loss = _compute_data_loss(series_dense_view_extended,
                                  from_t = from_t,
                                  to_t = to_t,
                                  sampling_interval = point_validity,
                                  force = force_compute_data_loss)
    logger.debug('Computed data loss: "{}"'.format(data_loss))

    # For each point, attach the "weight"/"contirbution
    for point in series_dense_view_extended:

        # Weight zero if completely outside the interval, which is required if in the operation
        # there is the need of knowing the next(or prev) value, even if it was far away
        if point.valid_to < from_t:
            point.weight = 0
            continue
        if point.valid_from >= to_t:
            point.weight = 0
            continue

        # Set this point valid from/to accoring to the interval boundaries
        this_point_valid_from = point.valid_from if point.valid_from >= from_t else from_t
        this_point_valid_to = point.valid_to if point.valid_to < to_t else to_t

        # Set weigth
        point.weight = (this_point_valid_to-this_point_valid_from)/interval_duration

        # Log
        logger.debug('Set series point @ %s weight: %s (data: %s)', point.dt, point.weight, point.data)

    # If creating slots, create also the slice of the series series containing only the slot datapoints
    if target == 'slot':

        if slot_first_point_i is None and slot_last_point_i is None:
            # If we have no datapoints at all in the slot, we must create one to let operations not
            # supporting weights to properly work. If we have a full data loss, this can be just the
            # middle point from the series_dense_view_extended. Otherwise, we must compute it from
            # scratch as we are upsamping with no data loss but still no points belonging to the slot.

            if data_loss==1:
                series_dense_view = None
                for point in series_dense_view_extended:
                    try:
                        if point._interpolated:
                            if series_dense_view:
                                raise ConsistencyException('Found more than one reconstructed point in a fully missing slot')
                            series_dense_view = series.__class__(point)
                    except AttributeError:
                        pass
                if not series_dense_view:
                    raise ConsistencyException('Could not find any reconstructed point in a fully missing slot')
            else:
                # Just create a point based on the series_dense_view_extended weights (TODO: are we including the interpolated?):
                new_point_data = {data_label:0 for data_label in data_labels}
                for point in series_dense_view_extended:
                    for data_label in data_labels:
                        point.data[data_label] += point.data[data_label] * point.weight
                new_point_t = (to_t - from_t) /2
                series_dense_view = TimeSeries(DataTimePoint(t=new_point_t, data=new_point_data, tz=series.tz))

        else:
            # Create a view of the original series to provide only the datapoints belonging to the slot
            #logger.critical('Slicing dense series from {} to {}'.format(slot_first_point_i, slot_last_point_i+1))
            series_dense_view = _TimeSeriesDenseView(series, from_i=slot_first_point_i, to_i=slot_last_point_i+1, # Slicing exclude the right
                                                     from_t=from_t, to_t=to_t, dense=True, interpolator_class=interpolator_class)



    # Compute point data
    if target=='point':

        # Compute the (weighted) average of all point contributions
        avgs = avg(series_dense_view_extended)

        # ...and assign them to the data value
        if isinstance(avgs, dict):
            data = {data_label:avgs[data_label] for data_label in data_labels}
        else:
            data = {data_labels[0]: avgs}

    #  Compute slot data
    elif target=='slot':

        if data_loss == 1 and fill_with:
            for data_label in data_labels:
                for operation in operations:
                    data['{}_{}'.format(data_label, operation.__name__)] = fill_with

        else:
            # Handle operations
            for operation in operations:
                try:
                    operation_supports_weights = operation._supports_weights
                except AttributeError:
                    # This is because user-defined operations can be even simple functions
                    # or based on custom classes without the _supports_weights attribute
                    operation_supports_weights = False

                if operation_supports_weights:
                    operation_data = operation(series_dense_view_extended)
                else:
                    operation_data = operation(series_dense_view)

                # ...and add to the data
                if isinstance(operation_data, dict):
                    for result_data_label in operation_data:
                        data['{}_{}'.format(result_data_label, operation.__name__)] = operation_data[result_data_label]
                else:
                    data['{}_{}'.format(data_labels[0], operation.__name__)] = operation_data

        if not data:
            raise Exception('No data computed at all?')

    else:
        raise ValueError('No idea how to compute a new "{}"'.format(target))


    # Do we have a force data_loss?
    #TODO: do not compute data_loss if fill_with not present and force_data_loss
    if force_data_loss is not None:
        data_loss = force_data_loss

    # Create the new item
    if target=='point':
        new_element = DataTimePoint(t = (from_t+((interval_duration)/2)),
                                    tz = timezone,
                                    data  = data,
                                    data_loss = data_loss)
    elif target=='slot':
        # Create the DataTimeSlot
        new_element = DataTimeSlot(start = TimePoint(t=from_t, tz=timezone),
                                   end   = TimePoint(t=to_t, tz=timezone),
                                   unit  = unit,
                                   data  = data,
                                   data_loss = data_loss)
    else:
        raise ValueError('No idea how to create a new "{}"'.format(target))

    # Handle data_indexes TODO: check math and everything here
    for index in series_data_indexes:

        # Skip the data loss as it is recomputed with different logics
        if index == 'data_loss':
            continue

        if series_dense_view_extended:
            index_total = 0
            index_total_weights = 0
            for point in series_dense_view_extended:

                # Get index value
                try:
                    index_value = point.data_indexes[index]
                except:
                    pass
                else:
                    if index_value is not None:
                        index_total_weights += point.weight
                        index_total += index_value * point.weight
                        #logger.critical('%s@%s: %s * %s', index, point.dt, index_value, point.weight)

            # Compute the new index value (if there were data_indexes not None)
            if index_total_weights > 0:

                # Project (rescale/normalize), because we could have some points without the index at all
                new_element_index_value = index_total/index_total_weights
                #logger.debug('%s/%s', index_total, index_total_weights)

            else:
                new_element_index_value = None

        else:
            new_element_index_value = None

        # Set the index in the data indexes
        new_element.data_indexes[index] = new_element_index_value

    # Return
    return new_element


#==========================
#  Base Transformation
#==========================

class Transformation(object):
    """A generic transformation."""

    @classmethod
    def __str__(cls):
        return '{} transformation'.format(cls.__name__.replace('Transformation',''))

    def process(self, series, target, start=None, end=None, validity=None, include_extremes=False,
                fill_with=None,  force_data_loss=None, fill_gaps=True, **kwargs):
        """Start the transformation process. If start and/or end are not set, they are set automatically
        based on first and last points of the series"""

        if not isinstance(series, TimeSeries):
            raise NotImplementedError('Transformations work only with TimeSeries data for now (got "{}")'.format(series.__class__.__name__))

        if not (issubclass(series.item_type, Point) or issubclass(series.item_type, Slot)):
                raise TypeError('Series items are not Points nor Slots, cannot compute any transformation')

        # Checks
        if include_extremes:
            raise NotImplementedError('Including the extremes is not yet implemented')
        if not issubclass(series.item_type, DataTimePoint):
            raise TypeError('Can process only time series of DataTimePoints, got time series of items type "{}"'.format(series.item_type.__name__))
        if not series:
            raise ValueError('Cannot process empty series')
        if target not in ['points', 'slots']:
            raise ValueError('Don\'t know how to target "{}"'.format(target))

        # Log
        logger.debug('Computing from/to with include_extremes= %s', include_extremes)

        # Handle start/end. If creating points, we will move the start back of half unit
        # and the end forward of half unit as well, as the point will be in the center
        from_t = None
        to_t = None
        from_dt = None
        to_dt = None
        if start is not None:
            if isinstance(start, datetime):
                from_dt = start
            else:
                try:
                    from_t = float(start)
                except:
                    raise ValueError('Cannot use "{}" as start value, not a datetime nor an epoch timestamp'.format(start))
        if end is not None:
            if isinstance(end, datetime):
                to_dt = end
            else:
                try:
                    to_t = float(end)
                except:
                    raise ValueError('Cannot use "{}" as end value, not a datetime nor an epoch timestamp'.format(end))

        # Set "from". TODO: check against shifted rounding if points?
        if from_t:
            from_dt = dt_from_s(from_t)
            if from_dt != self.time_unit.round(from_dt):
                raise ValueError('The provided from_t is not consistent with the self.time_unit of "{}" (Got "{}")'.format(self.time_unit, from_t))
        elif from_dt:
            from_t = s_from_dt(from_dt)
            if from_dt != self.time_unit.round(from_dt):
                raise ValueError('The provided from_dt is not consistent with the self.time_unit of "{}" (Got "{}")'.format(self.time_unit, from_dt))
        else:
            if target == 'points':
                from_t = series[0].t  - (self.time_unit.as_seconds() /2)
                from_dt = dt_from_s(from_t, tz=series.tz)
                from_dt = self.time_unit.round(from_dt, how='floor' if include_extremes else 'ceil')
                from_t = s_from_dt(from_dt) + (self.time_unit.as_seconds() /2)
                from_dt = dt_from_s(from_t)

            elif target == 'slots':
                from_t = series[0].t
                from_dt = dt_from_s(from_t, tz=series.tz)
                if from_dt == self.time_unit.round(from_dt):
                    # Only for the start, and only for the slots, if the first point is
                    # exactly equal to the first slot start, leave it as it is to include it.
                    pass
                else:
                    from_dt = self.time_unit.round(from_dt, how='floor' if include_extremes else 'ceil')
                from_t = s_from_dt(from_dt)
                from_dt = dt_from_s(from_t)
            else:
                raise ValueError('Don\'t know how to target "{}"'.format(target))

        # Set "to". TODO: check against shifted rounding if points?
        if to_t:
            to_dt = dt_from_s(to_t)
            if to_dt != self.time_unit.round(to_dt):
                raise ValueError('The provided to_t is not consistent with the self.time_unit of "{}" (Got "{}")'.format(self.time_unit, to_t))
        elif to_dt:
            to_t = s_from_dt(to_dt)
            if to_dt != self.time_unit.round(to_dt):
                raise ValueError('The provided to_dt is not consistent with the self.time_unit of "{}" (Got "{}")'.format(self.time_unit, to_dt))
        else:
            if target == 'points':
                to_t = series[-1].t  + (self.time_unit.as_seconds() /2)
                to_dt = dt_from_s(to_t, tz=series.tz)
                from_dt = self.time_unit.round(from_dt, how='ceil' if include_extremes else 'floor')
                to_t = s_from_dt(to_dt) - (self.time_unit.as_seconds() /2)
                to_dt = dt_from_s(to_t)
            elif target == 'slots':
                to_t = series[-1].t
                to_dt = dt_from_s(to_t, tz=series.tz)
                from_dt = self.time_unit.round(from_dt, how='ceil' if include_extremes else 'floor')
                to_t = s_from_dt(to_dt)
                to_dt = dt_from_s(to_t)
            else:
                raise ValueError('Don\'t know how to target "{}"'.format(target))

        # Set/fix timezone
        from_dt = as_tz(from_dt, series.tz)
        to_dt = as_tz(to_dt, series.tz)

        # Log
        logger.debug('Computed from: %s', from_dt)
        logger.debug('Computed to: %s',to_dt)

        # Automatically detect validity if not set
        if validity is None:
            validity = series._autodetected_sampling_interval
            logger.info('Using auto-detected sampling interval: %ss', validity)

        # Check if upsamplimg (with some tolearance):
        if validity > (self.time_unit.as_seconds(series[0].dt) * 1.10):
            logger.warning('You are upsampling, which is not well tested yet. Expect potential issues.')

        # Compute validity regions
        validity_regions = _compute_validity_regions(series, sampling_interval=validity)

        # Log
        logger.debug('Started resampling from "%s" (%s) to "%s" (%s)', from_dt, from_t, to_dt, to_t)

        # Set some support vars
        slot_start_t = None
        slot_start_dt = None
        slot_end_t = None
        slot_end_dt = None
        process_ended = False
        slot_prev_point_i = None
        new_series = TimeSeries()

        # Set timezone
        timezone  = series.tz
        logger.debug('Using timezone "%s"', timezone)

        # Counters
        count = 0
        first = True

        # data_indexes & resolution shortcuts
        series_data_indexes = series._all_data_indexes()
        series_resolution = series.resolution

        # Set operations if slots
        if target == 'slots':
            operations  = self.operations
        else:
            operations = None

        # Now go trough all the data in the time series
        for i, point in enumerate(series):

            logger.debug('Processing i=%s: %s', i, point)

            # Increase counter
            count += 1

            # Attach validity TODO: do we want a generic series method, or maybe to be always computed in the series?
            point.valid_from = validity_regions[point.t][0]
            point.valid_to = validity_regions[point.t][1]
            logger.debug('Attached validity to %s', point.t)

            # Pretend there was a slot before if we are at the beginning. TOOD: improve me.
            if slot_end_t is None:
                slot_end_t = from_t

            # First, check if we have some points to discard at the beginning
            if point.t < from_t:
                # If we are here it means we are going data belonging to a previous slot
                logger.debug('Discarding as before start')
                # NOTE: here we set the prev index
                slot_prev_point_i = i
                continue

            # Similar concept for the end
            # TODO: what if we are in streaming mode? add if to_t is not None?
            if point.t >= to_t:
                if process_ended:
                    logger.debug('Discarding as after end')
                    continue

            # Is the current slot outdated? (are we processing a datapoint falling outside the current slot?)
            if point.t >= slot_end_t:

                # This approach is to detect if the current slot is "outdated" and spin a new one if so.
                # Works only for slots at the beginning and in the middle, but not for the last slot
                # or the missing slots at the end which need to be closed down here
                logger.debug('Detected outdated slot')

                # Keep spinning new slots until the current data point falls in one of them.
                # NOTE: Read the following "while" more as an "if" which can also lead to spin multiple
                # slot if there are empty slots between the one being closed and the point.dt.
                # TODO: leave or remove the above if for code readability?

                while slot_end_t <= point.t:
                    logger.debug('Checking for end %s with point %s', slot_end_t, point.t)

                    if slot_start_t is None:
                        # If we are in the pre-first slot, just silently spin a new slot
                        pass

                    else:

                        # Set slot next point index (us, as we triggered a new slot)
                        slot_next_point_i = i

                        # The slot_prev_point_i is instead always defined in the process and correctly set.
                        slot_prev_point_i = slot_prev_point_i

                        logger.debug('Got slot_prev_point_i as %s',slot_prev_point_i)
                        logger.debug('Got slot_next_point_i as %s',slot_next_point_i)

                        # Now set slot first and last point, but beware boundaries.
                        if slot_prev_point_i is not None and abs(slot_next_point_i - slot_prev_point_i) == 1:

                            # No points for the slot at all
                            slot_first_point_i = None
                            slot_last_point_i = None

                        else:

                            # Set first
                            if slot_prev_point_i is None:
                                # Edge case where there is no prev as the time series starts with a point
                                # placed exactly on the slot start, and since lefts are included we take it.
                                slot_first_point_i = 0
                            else:
                                if series[slot_prev_point_i+1].t >= slot_start_t and series[slot_prev_point_i+1].t < slot_end_t:
                                    slot_first_point_i = slot_prev_point_i+1
                                else:
                                    slot_first_point_i = None

                            # Set last
                            if series[slot_next_point_i-1].t >= slot_start_t and series[slot_next_point_i-1].t < slot_end_t:
                                slot_last_point_i = slot_next_point_i-1
                            else:
                                slot_last_point_i = None


                        logger.debug('Set slot_first_point_i to %s',slot_first_point_i)
                        logger.debug('Set slot_last_point_i to %s',slot_last_point_i)



                        # Log the new slot
                        # slot_first_point_i-1 is the "prev"
                        # slot_last_point_i+1 is the "next" (and the index where we are at the moment)
                        logger.debug('This slot is closed: start=%s (%s) and end=%s (%s). Now computing it..', slot_start_t, slot_start_dt, slot_end_t, slot_end_dt)

                        # Compute the new item...
                        new_item = _compute_new(target = 'point' if target=='points' else 'slot',
                                                series = series,
                                                slot_first_point_i = slot_first_point_i,
                                                slot_last_point_i = slot_last_point_i ,
                                                slot_prev_point_i = slot_prev_point_i,
                                                slot_next_point_i = slot_next_point_i,
                                                from_t = slot_start_t,
                                                to_t = slot_end_t,
                                                unit = self.time_unit,
                                                point_validity = validity,
                                                timezone = timezone,
                                                fill_with = fill_with,
                                                force_data_loss = force_data_loss,
                                                fill_gaps = fill_gaps,
                                                series_data_indexes = series_data_indexes,
                                                series_resolution = series_resolution,
                                                force_compute_data_loss = True if first else False,
                                                interpolator_class=self.interpolator_class,
                                                operations = operations)

                        # Set first to false
                        if first:
                            first = False

                        # .. and append results
                        if new_item:
                            logger.debug('Computed new item: %s',new_item )
                            new_series.append(new_item)

                    # Create a new slot. This is where all the "conventional" time logic kicks-in, and where the timezone is required.
                    slot_start_t = slot_end_t
                    slot_start_dt = dt_from_s(slot_start_t, tz=timezone)
                    # TODO: if not calendar time unit, we can just use seconds and maybe speed up a bit.
                    slot_end_dt = slot_start_dt + self.time_unit
                    slot_end_t = s_from_dt(slot_end_dt)

                    # Reassign prev (if we have to)
                    if i>0 and series[i-1].t < slot_start_t:
                        slot_prev_point_i = i-1

                    # Log the newly spinned slot
                    logger.debug('Spinned a new slot, start=%s (%s), end=%s (%s)', slot_start_t, slot_start_dt, slot_end_t, slot_end_dt)

                # If last slot mark process as completed
                if point.dt >= to_dt:
                    process_ended = True

        if target == 'points':
            logger.info('Resampled %s DataTimePoints in %s DataTimePoints', count, len(new_series))
        else:
            if isinstance(series[0], DataTimePoint):
                logger.info('Aggregated %s points in %s slots', count, len(new_series))
            else:
                logger.info('Aggregated %s slots in %s slots', count, len(new_series))

        return new_series


#==========================
#  Resampler
#==========================

class Resampler(Transformation):
    """Resampling transformation."""

    def __init__(self, unit, interpolator_class=LinearInterpolator):

        # Handle unit
        if isinstance(unit, TimeUnit):
            self.time_unit = unit
        else:
            self.time_unit = TimeUnit(unit)

        unit_type = ''.join([char for char in str(self.time_unit) if not char.isdigit()])
        if unit_type in ['Y', 'M', 'D']:
            raise ValueError('Sorry, time units involving calendar components are not supported by the Resampler (got "{}"). Use the Slotter instead.'.format(self.time_unit))

        # Set interpolator
        self.interpolator_class=interpolator_class

    def process(self, series, *args, **kwargs):
        kwargs['target'] = 'points'
        return super(Resampler, self).process(series, *args, **kwargs)


#==========================
#   Aggregator
#==========================

class Aggregator(Transformation):
    """Aggregation transformation."""

    def __init__(self, unit, operations=[avg], interpolator_class=LinearInterpolator):

        # Handle unit
        if isinstance(unit, TimeUnit):
            self.time_unit = unit
        else:
            self.time_unit = TimeUnit(unit)

        if not operations:
            raise ValueError('No operations set, cannot slot.')

        # Handle operations
        operations_input = operations
        operations = []
        for operation in operations_input:
            if isinstance(operation, str):
                try:
                    operation = getattr(operations_module, operation)
                except:
                    raise
            operations.append(operation)
        self.operations = operations

        # Set interpolator
        self.interpolator_class=interpolator_class

    def process(self, series, *args, **kwargs):
        kwargs['target'] = 'slots'
        return super(Aggregator, self).process(series, *args, **kwargs)







# -*- coding: utf-8 -*-
"""Base model classes."""

import os
import json
import uuid
import functools
import statistics
from propertime.utils import now_s, dt_from_s, s_from_dt
from datetime import datetime
from pandas import DataFrame
import shutil
import copy

from ..exceptions import NotFittedError, AlreadyFittedError
from ..utils import _check_timeseries, _check_resolution, _check_data_labels, _item_is_in_range
from ..units import TimeUnit
from ..datastructures import Point, Series, TimeSeries, TimeSeriesView

# Setup logging
import logging
logger = logging.getLogger(__name__)

# Suppress TensorFlow warnings as default behavior
try:
    import tensorflow as tf
    tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
except (ImportError,AttributeError):
    pass


class TooMuchDataLoss(Exception):
    pass

#======================
#  Base Model
#======================

class Model():
    """A generic model. This can be either a stateless model, where all the information is coded and there are no parameters, or a
    stateful (parametric) model, where there are a number of parameters which can be both set manually or learnt (fitted) from the data.

    All models expose a ``predict()``, ``apply()`` and ``evaluate()`` methods, while parametric models also provide a ``save()`` method to
    store the parameters of the model, and an optional ``fit()`` method if the parameters are to be learnt form data. In this case also a
    ``cross_validate()`` method is available.

    Series resolution and data labels consistency are enforced between all methods and save/load operations.

    Args:
        path (optional, str): a path from which to load a saved model. Will override all other init settings. Only for parametric models.
    """

    def __init__(self):

        # Set type
        try:
            self.data
        except:
            # No data (parameters) for the model. Check if there is a fit implemented
            if self._is_fit_implemented():
                # Yes: the model is parameter and the parameters
                # are still to be set via the fit
                self._type = 'parametric'
            else:
                # No: the model has no parameters at all
                self._type = 'non-parametric'
        else:
            # If we have data, it means that there are parameters and
            # the model is parametric.
            self._type = 'parametric'

        # Set the model ID
        id = str(uuid.uuid4())
        self.fitted = False
        try:
            self.data['id'] = id
        except AttributeError:
            self.data = {'id': id}

    @property
    def id(self):
        """A unique identifier for the model. Only for parametric models."""
        if not self._is_parametric():
            raise TypeError('Non-parametric models have no ID') # TODO: is this the right exception?
        else:
            return self.data['id']

    @classmethod
    def load(cls, path):
        """Load the model from a path."""
        # Do we have to load the model (if parametric?)
        #if not cls._is_parametric():
        #    raise ValueError('Loading a non-parametric model from a path does not make sense')

        # Create the model
        model = cls()

        # Set the data (replaces the model ID also)
        with open(path+'/data.json', 'r') as f:
            model.data = json.loads(f.read())

        # TODO: the "fitted" does not apply for models with parameters but with no fit.. fix me.
        model.fitted = True

        # Convert resolution to TimeUnit if any. TODO: right now only TimeSeries have the resolution,
        # but might change in case of adding the resolution attribute also to generic series.
        if 'resolution' in model.data:
            model.data['resolution'] = TimeUnit(model.data['resolution'])

        return model

    def save(self, path):
        """Save the model in the given path. The model is saved in "directory format",
         meaning that a new directory will be created containing the data for the model."""

        # Is this model parametric?
        if not self._is_parametric():
            raise TypeError('Saving a non-parametric model from a path does not make sense') # TODO: is this the right exception?

        # Ensure the model is fitted if it has to
        if self._is_fit_implemented() and not self.fitted:
            raise NotFittedError()

        if not path:
            raise ValueError('Got empty path, cannot save')

        if os.path.exists(path):
            raise ValueError('The path "{}" already exists'.format(path))

        # Prepare model dir and dump data as json
        os.makedirs(path)
        model_data_file = '{}/data.json'.format(path)

        # Temporary change the resolution to its string representation (if any)
        if 'resolution' in self.data:
            resolution_obj = self.data['resolution']
            self.data['resolution'] = str(resolution_obj)

        # Write the data
        try:
            with open(model_data_file, 'w') as f:
                f.write(json.dumps(self.data))
        finally:
            # In any case revert resolution back to object (if any)
            if 'resolution' in self.data:
                self.data['resolution'] = resolution_obj

        logger.info('Saved model with id "%s" in "%s"', self.data['id'], path)


    def _is_fit_implemented(self):
        """If the model has a fit method implemented or not."""
        try:
            self.fit(series=None)
        except NotImplementedError:
            return False
        except Exception:
            return True
        else:
            return True

    def _is_parametric(self):
        """If the model is parametric or not."""
        if self._type == 'parametric':
            return True
        else:
            return False

    @staticmethod
    def fit_method(fit_method):
        """:meta private:"""
        @functools.wraps(fit_method)
        def do_fit(self, series, *args, **kwargs):

            if self.fitted:
                raise AlreadyFittedError('This model is already fitted. Use the fit_update() method if you want to update it with '
                                         + 'new data (where implemented). If you instead want to re-fit it, create a new instance.')

            # Check data
            if not isinstance(series, TimeSeries):
                raise TypeError('Models work only with TimeSeries data for now (got "{}")'.format(series.__class__.__name__))

            # If TimeSeries data, check it
            if isinstance(series, TimeSeries):
                _check_timeseries(series)

                # Set resolution
                try:
                    self.data['resolution'] = series.resolution
                except AttributeError:
                    pass

                # Set data labels
                try:
                    self.data['data_labels'] = series.data_labels()
                except AttributeError:
                    pass

            # Call fit logic
            fit_output = fit_method(self, series, *args, **kwargs)

            self.data['fitted_at'] = now_s()
            self.fitted = True

            return fit_output

        return do_fit

    def fit(self, series, verbose=False):
        """Fit the model on a series."""
        raise NotImplementedError('Fitting this model is not implemented')


    @staticmethod
    def fit_update_method(fit_update_method):
        """:meta private:"""
        @functools.wraps(fit_update_method)
        def do_fit_update(self, series, *args, **kwargs):

            if not self.fitted:
                raise NotFittedError('Cannot update the fit for a non-fitted model')

            # Check data
            if not isinstance(series, TimeSeries):
                raise TypeError('Models work only with TimeSeries data for now (got "{}")'.format(series.__class__.__name__))

            # If TimeSeries data, check it
            if isinstance(series, TimeSeries):
                _check_timeseries(series)

            # Check resolution
            if 'resolution' in self.data:
                _check_resolution(series, self.data['resolution'])

            # Check data labels
            if 'data_labels' in self.data:
                _check_data_labels(series, self.data['data_labels'])

            # Call update fit logic
            fit_update_output = fit_update_method(self, series, *args, **kwargs)

            # Update fitted at and model id
            self.data['fitted_at'] = now_s()
            self.data['id'] = str(uuid.uuid4())

            return fit_update_output

        return do_fit_update

    def fit_update(self, series, verbose=False, **kwargs):
        """Update the model fit on a series."""
        raise NotImplementedError('Updating the fit is not implemented for this model')

    @staticmethod
    def predict_method(predict_method):
        """:meta private:"""
        @functools.wraps(predict_method)
        def do_predict(self, series, *args, **kwargs):

            # Ensure the model is fitted if it has to.
            if self._is_fit_implemented() and not self.fitted:
                raise NotFittedError()

            # Check data
            if not isinstance(series, TimeSeries):
                raise TypeError('Models work only with TimeSeries data for now (got "{}")'.format(series.__class__.__name__))

            # If TimeSeries data, check it
            if isinstance(series, TimeSeries):
                _check_timeseries(series)
                if self._is_parametric():
                    if isinstance(series.item_type, Point) and len(series) == 1:
                        # Do not check if the data is a point time series and has only one item
                        pass
                    else:
                        _check_resolution(series, self.data['resolution'])
                    _check_data_labels(series, self.data['data_labels'])

            # Check context
            if 'context_data' in kwargs and kwargs['context_data'] and ('context_data_labels' not in self.data or not self.data['context_data_labels']):
                raise ValueError('This model does not accept context data')
            if 'context_data_labels' in self.data and self.data['context_data_labels'] and ('context_data' not in kwargs or not kwargs['context_data']):
                raise ValueError('This model requires context data ({})'.format(self.data['context_data_labels']))

            # Call predict logic
            return predict_method(self, series, *args, **kwargs)

        return do_predict

    def predict(self, series):
        """Call the model predict logic on a series."""
        raise NotImplementedError('Predicting with this model is not implemented')

    @staticmethod
    def apply_method(apply_method):
        """:meta private:"""
        @functools.wraps(apply_method)
        def do_apply(self, series, *args, **kwargs):

            # Ensure the model is fitted if it has to.
            if self._is_fit_implemented() and not self.fitted:
                raise NotFittedError()

            # Check data
            if not isinstance(series, TimeSeries):
                raise TypeError('Models work only with TimeSeries data for now (got "{}")'.format(series.__class__.__name__))

            # If TimeSeries data, check it
            if isinstance(series, TimeSeries):
                _check_timeseries(series)
                if self._is_parametric():
                    if isinstance(series.item_type, Point) and len(series) == 1:
                        # Do not check if the data is a point time series and has only one item
                        pass
                    else:
                        _check_resolution(series, self.data['resolution'])
                    _check_data_labels(series, self.data['data_labels'])

            # Call apply logic
            return apply_method(self, series, *args, **kwargs)

        return do_apply

    def apply(self, series):
        """Call the model apply logic on a series."""
        raise NotImplementedError('Applying this model is not implemented')

    @staticmethod
    def evaluate_method(evaluate_method):
        """:meta private:"""
        @functools.wraps(evaluate_method)
        def do_evaluate(self, series, *args, **kwargs):

            # Ensure the model is fitted if it has to.
            if self._is_fit_implemented() and not self.fitted:
                raise NotFittedError()

            # Check data
            if not isinstance(series, TimeSeries):
                raise TypeError('Models work only with TimeSeries data for now (got "{}")'.format(series.__class__.__name__))

            # If TimeSeries data, check it
            if isinstance(series, TimeSeries):
                _check_timeseries(series)
                if self._is_parametric():
                    if isinstance(series.item_type, Point) and len(series) == 1:
                        # Do not check if the data is a point time series and has only one item
                        pass
                    else:
                        _check_resolution(series, self.data['resolution'])
                    _check_data_labels(series, self.data['data_labels'])

            # Call evaluate logic
            return evaluate_method(self, series, *args, **kwargs)

        return do_evaluate

    def evaluate(self, series):
        """Call the model evaluate logic on a series."""
        raise NotImplementedError('Evaluating this model is not implemented')

    def cross_validate(self, series, rounds=10, return_full_evaluations=False, **kwargs):
        """Cross validate the model on a series, by default with 10 fit/evaluate rounds.

        All the parameters starting with the ``fit_`` prefix are forwarded to the model ``fit()`` method (without the prefix), and
        all the parameters starting with the ``evaluate_`` prefix are forwarded to the model ``evaluate()`` method (without the prefix).

        Args:
            rounds(int): how many rounds of cross validation to run.
            return_full_evaluations(bool): if to return the full evaluations, one for each round.
        """

        if self.fitted:
            raise NotImplementedError('You are trying to cross-validate a model already fitted, this is not supported.')

        # Check data
        if not isinstance(series, TimeSeries):
            raise TypeError('Models work only with TimeSeries data for now (got "{}")'.format(series.__class__.__name__))

        # If TimeSeries data, check it
        if isinstance(series, TimeSeries):
            _check_timeseries(series)

        # Decouple evaluate from fit args
        fit_kwargs = {}
        evaluate_kwargs = {}
        consumed_kwargs = []
        for kwarg in kwargs:
            if kwarg.startswith('fit_'):
                consumed_kwargs.append(kwarg)
                fit_kwargs[kwarg.replace('fit_', '')] = kwargs[kwarg]
            if kwarg.startswith('evaluate_'):
                consumed_kwargs.append(kwarg)
                evaluate_kwargs[kwarg.replace('evaluate_', '')] = kwargs[kwarg]
        for consumed_kwarg in consumed_kwargs:
            kwargs.pop(consumed_kwarg)

        # Do we still have some kwargs?
        if kwargs:
            raise Exception('Got some unknown args: {}'.format(kwargs))

        # How many items per round?
        round_items = int(len(series) / rounds)
        logger.debug('Items per round: {}'.format(round_items))

        # Start the fit / evaluate loop
        evaluations = []
        fit_update_unaivalable_warned = False
        for i in range(rounds):
            validate_from_i = round_items*i
            validate_to_i =   (round_items*i)+round_items
            validate_from_t = series[validate_from_i].t
            try:
                validate_to_t = series[validate_to_i].t
            except IndexError:
                validate_to_t = series[validate_to_i - 1].t
            validate_from_dt = dt_from_s(validate_from_t)
            validate_to_dt   = dt_from_s(validate_to_t)

            # Deep-copy the model
            model = copy.deepcopy(self)

            # Perform the cross validation
            logger.info('Cross validation round {}/{}: validate from {} ({}) to {} ({}), fit on the rest.'.format(i+1, rounds, validate_from_t, validate_from_dt, validate_to_t, validate_to_dt))

            if validate_from_i == 0:
                # First chunk
                logger.debug('Fitting from {} to the end'.format(validate_to_t))
                series_view = TimeSeriesView(series=series, from_i=validate_to_i, to_i=len(series))
                model.fit(series_view, **fit_kwargs)

            elif validate_to_i >= len(series):
                # Last chunk
                logger.debug('Fitting from the beginning to {}'.format(validate_from_t))
                series_view = TimeSeriesView(series=series, from_i=0, to_i=validate_from_i)
                model.fit(series_view, **fit_kwargs)

            else:
                # Find the bigger chunk and fit on that
                if validate_from_i > len(series)-validate_to_i:
                    logger.debug('Fitting from the beginning to {}'.format(validate_from_t))
                    series_view = TimeSeriesView(series=series, from_i=0, to_i=validate_from_i)
                    model.fit(series_view, **fit_kwargs)
                    model_window = None 
                    try:
                        model_window = model.window
                    except AttributeError:
                        pass
                    if model_window and (len(series)-validate_to_i) < model_window:
                        pass
                    else:
                        # Try to fit on the other chunk as well:
                        try:
                            logger.debug('Now trying to fit also from {} to the end'.format(validate_to_t))
                            series_view = TimeSeriesView(series=series, from_i=validate_to_i, to_i=len(series))
                            model.fit_update(series_view, **fit_kwargs)
                        except NotImplementedError:
                            if not fit_update_unaivalable_warned:
                                logger.warning('This model does not support updating the fit, cross validation results will be approximate for the intermediate chunks')
                                fit_update_unaivalable_warned = True
                else:
                    logger.debug('Fitting from {} to the end'.format(validate_to_t))
                    series_view = TimeSeriesView(series=series, from_i=validate_to_i, to_i=len(series))
                    model.fit(series_view, **fit_kwargs)
                    model_window = None 
                    try:
                        model_window = model.window
                    except AttributeError:
                        pass
                    if model_window and validate_from_i < model_window:
                        pass
                    else:
                        # Try to fit on the other chunk as well:
                        try:
                            logger.debug('Now trying to fit also from the beginning to {}'.format(validate_from_t))
                            series_view = TimeSeriesView(series=series, from_i=0, to_i=validate_from_i)
                            model.fit_update(series_view, **fit_kwargs)
                        except NotImplementedError:
                            if not fit_update_unaivalable_warned:
                                logger.warning('This model does not support updating the fit, cross validation results will be approximate for the intermediate chunks')
                                fit_update_unaivalable_warned = True

            # Evaluate & append
            evaluations.append(model.evaluate(TimeSeriesView(series=series, from_i=validate_from_i, to_i=validate_to_i), **evaluate_kwargs))

        # Regroup evaluations
        evaluation_metrics = list(evaluations[0].keys())
        scores_by_evaluation_metric = {}
        for evaluation in evaluations:
            for evaluation_metric in evaluation_metrics:
                try:
                    scores_by_evaluation_metric[evaluation_metric].append(evaluation[evaluation_metric])
                except KeyError:
                    try:
                        scores_by_evaluation_metric[evaluation_metric] = [evaluation[evaluation_metric]]
                    except KeyError:
                        raise Exception('Error, the model generated different evaluation metrics over the rounds, cannot compute cross validation.') from None

        # Prepare and return results
        results = {}
        for evaluation_metric in scores_by_evaluation_metric:
            results[evaluation_metric+'_avg'] = statistics.mean(scores_by_evaluation_metric[evaluation_metric])
            results[evaluation_metric+'_stdev'] = statistics.stdev(scores_by_evaluation_metric[evaluation_metric])
        if return_full_evaluations:
            results['evaluations'] = evaluations
        return results


#=========================
#  Base Prophet model
#=========================

class _ProphetModel(Model):
    '''A model using Prophet as underlying engine, and providing some extra internal functions for common operations.'''

    @classmethod
    def _remove_timezone(cls, dt):
        return dt.replace(tzinfo=None)

    @classmethod
    def _from_timeseria_to_prophet(cls, series):

        # Create Python lists with data
        try:
            series[0].data[0]
            data_labels_are_indexes = True
        except KeyError:
            series[0].data.keys()
            data_labels_are_indexes = False

        data_as_list=[]
        for item in series:

            if data_labels_are_indexes:
                data_as_list.append([cls._remove_timezone(item.dt), item.data[0]])
            else:
                data_as_list.append([cls._remove_timezone(item.dt), item.data[list(item.data.keys())[0]]])

        # Create the pandas DataFrames
        data = DataFrame(data_as_list, columns = ['ds', 'y'])

        return data


#=========================
#  Base ARIMA model
#=========================

class _ARIMAModel(Model):
    '''A model using statsmodel's ARIMA as underlying engine, and providing some extra internal functions for common operations.'''

    def _get_start_end_indexes(self, series, n):

        # Do the math to get the right indexes for the prediction, both out-of-sample and in-sample
        # Not used at the moment as there seems to be bugs in the statsmodel package.
        # See https://www.statsmodels.org/devel/generated/statsmodels.tsa.arima_model.ARIMAResults.predict.html

        # The default start_index for an ARIMA prediction is the index after the fit series
        # I.e. if the fit series had 101 elments, the start index is the 101
        # So we need to compute the start and end indexes according to the fit series,
        # we will use the "resolution" for this.

        # Save the requested prediction "from"
        requested_predicting_from_dt = series[-1].dt + series.resolution

        # Search for the index number TODO: try first with "pure" math which would work for UTC
        slider_dt = self.fit_series[0].dt
        count = 0
        while True:
            if slider_dt  == requested_predicting_from_dt:
                break

            # To next item index
            if requested_predicting_from_dt < self.fit_series[0].dt:
                slider_dt = slider_dt - self.fit_series.resolution
                if slider_dt < requested_predicting_from_dt:
                    raise Exception('Miss!')

            else:
                slider_dt = slider_dt + self.fit_series.resolution
                if slider_dt > requested_predicting_from_dt:
                    raise Exception('Miss!')

            count += 1

        # TODO: better understand the indexes in statsmodels (and potentially their bugs), as sometimes
        # setting the same start and end gave two predictions which is just nonsense in any scenario.
        start_index = count
        end_index = count + n

        return (start_index, end_index)


#=========================
#  Base Keras model
#=========================

class _KerasModel(Model):
    '''A model using Keras as underlying engine, and providing some extra internal functions for common operations.'''

    def _load_keras_model(self, path):

        # Load the Keras model
        if path:
            from tensorflow.keras.models import load_model as load_keras_model
            self.keras_model = load_keras_model('{}/model.keras'.format(path))

    def _save_keras_model(self, path):

        # Save the Keras model
        try:
            self.keras_model.save('{}/model.keras'.format(path))
        except Exception as e:
            shutil.rmtree(path)
            raise e

    # TODO: since we are extending a generic ParametricModel, the following methods should not be here.
    # Maybe only if extending a TimeSeriesParametricModel, and still it is probably the wrong place, as for the
    # ARIMA and Prophet above also. Consider moving them in a "utility" package or directly in the models.

    @staticmethod
    def _to_matrix_representation(series, window, steps, context_data_labels, target_data_labels, data_loss_limit):

        if steps > 1:
            raise NotImplementedError('Not implemented for steps >1')

        window_elements_matrix = []
        target_values_vector = []
        context_data_vector = []

        for i, _ in enumerate(series):
            if i <  window:
                continue
            if i == len(series):
                break

            try:
                # Add window elements
                window_elements_vector = []
                for j in range(window):
                    if data_loss_limit is not None and 'data_loss' in series[i-window+j].data_indexes and series[i-window+j].data_indexes['data_loss'] >= data_loss_limit:
                        raise TooMuchDataLoss()
                    window_elements_vector.append(series[i-window+j])

                # Add target values
                target_values_sub_vector = []
                if data_loss_limit is not None and 'data_loss' in series[i].data_indexes and series[i].data_indexes['data_loss'] >= data_loss_limit:
                    raise TooMuchDataLoss()
                for target_data_label in target_data_labels:
                    target_values_sub_vector.append(series[i].data[target_data_label])
            except TooMuchDataLoss:
                continue
            else:

                # Append window elements and target data
                window_elements_matrix.append(window_elements_vector)
                target_values_vector.append(target_values_sub_vector)

                # Add context data if required
                if context_data_labels is not None:
                    context_data_vector.append({data_label: series[i].data[data_label] for data_label in context_data_labels})

        return window_elements_matrix, target_values_vector, context_data_vector


    @staticmethod
    def _compute_window_features(window_datapoints, data_labels, time_unit, features, context_data=None, flatten=False):
        """Compute features from a list of window data points (or slots).

        Args:
            window_datapoints (list): The list with the data points (or slots)
            data_labels(dict): the labels of the point (or slot) data.
            features(list): the list of the features to compute.
                Supported values are:
                ``values`` (use the data values),
                ``diffs``  (use the diffs between the values), and
                ``hours``  (use the hours of the timestamp).
            flatten(bool): if to flatten the features as a signle list.
        """

        available_features = ['values', 'diffs', 'hours']
        for feature in features:
            if feature not in available_features:
                raise ValueError('Unknown feature "{}"'.format(feature))

        # Handle context data if any (as "fake", context datapoint).
        if context_data:
            context_datapoint_dt = window_datapoints[-1].dt + time_unit
            context_datapoint = window_datapoints[-1].__class__(dt = context_datapoint_dt, data = context_data)
            window_datapoints.append(context_datapoint)

        # Compute the features
        window_features = []
        for i in range(len(window_datapoints)):

            datapoint_features = []

            # 1) Data point/slot values (for all data labels)
            if 'values' in features:
                for data_label in data_labels:
                    #if window_datapoints[i].data[data_label] is not None:
                    try:
                        datapoint_features.append(window_datapoints[i].data[data_label])
                    except KeyError:
                        datapoint_features.append(0.0)

            # 2) Compute diffs on normalized datapoints
            if 'diffs' in features:
                for data_label in data_labels:
                    try:
                        if i == 0:
                            diff = window_datapoints[1].data[data_label] - window_datapoints[0].data[data_label]
                        elif i == len(window_datapoints)-1:
                            diff = window_datapoints[-1].data[data_label] - window_datapoints[-2].data[data_label]
                        else:
                            diff = (window_datapoints[i+1].data[data_label] - window_datapoints[i-1].data[data_label]) /2
                        if diff == 0:
                            diff = 1
                        datapoint_features.append(diff)
                    except KeyError:
                        datapoint_features.append(0.0)

            # 3) Hour (normlized)
            if 'hours' in features:
                datapoint_features.append(window_datapoints[i].dt.hour/24)

            # Now add to the window features
            if flatten:
                window_features += datapoint_features
            else:
                window_features.append(datapoint_features)

        return window_features





# -*- coding: utf-8 -*-
"""Base data structures as Points, Slots and Series."""

import json
from copy import deepcopy
from pandas import DataFrame, concat
from datetime import datetime
from pytz import UTC
from propertime import Time
from propertime.utils import s_from_dt , dt_from_s, timezonize, dt_from_str, str_from_dt

from .units import Unit, TimeUnit
from .utils import _is_close, _to_time_unit_string
from .utils import _is_index_based, _is_key_based, _has_numerical_values
from .exceptions import ConsistencyException

# Setup logging
import logging
logger = logging.getLogger(__name__)


#======================
#  Points
#======================

class Point():
    """A point.

       Args:
           *args (list): the coordinates.
    """

    def __init__(self, *args):
        if not args:
            raise Exception('A Point requires at least one coordinate, got none.')

        # Validate
        for arg in args:
            try:
                float(arg)
            except:
                raise Exception('Got non-numerical argument: "{}"'.format(arg))

        # Store (as tuple)
        self.coordinates = tuple(args)

    @property
    def __coordinates_repr__(self):
        if len(self.coordinates)==1:
            return str(self.coordinates[0])
        else:
            return str(self.coordinates).replace(' ', '')

    def __repr__(self):
        return 'Point @ {}'.format(self.__coordinates_repr__)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        try:
            if not self.coordinates == other.coordinates:
                return False
            for i in range(0, len(self.coordinates)):
                if self.coordinates[i] != other.coordinates[i]:
                    return False
            return True
        except (AttributeError, IndexError):
            return False

    def __gt__(self, other):
        if len(self.coordinates) > 1:
            raise ValueError('Comparing multi-dimensional points does not make sense')
        else:
            try:
                if len(other.coordinates) > 1:
                    raise ValueError('Comparing multi-dimensional points does not make sense')
                else:
                    return self.coordinates[0] > other.coordinates[0]
            except AttributeError:
                return self.coordinates[0] > other

    def __getitem__(self,index):
        return self.coordinates[index]


class TimePoint(Point):
    """A point in the time dimension. Can be initialized using the special `t` and `dt` arguments,
    for epoch seconds and datetime objects, respectively.

       Args:
           t (float): epoch timestamp, decimals for sub-second precision.
           dt (datetime): a datetime object timestamp.
    """

    def __init__(self, *args, **kwargs):

        # Handle timezone if any (removing it from kwargs)
        tz = kwargs.pop('tz', None)
        if tz:
            self._tz = timezonize(tz)

        # Do we have to handle kwargs?
        if kwargs:
            if 't' in kwargs:

                # Ok, will create the point in the standard way, just used the "t" kwarg to set the time coordinate
                t = kwargs['t']

            elif 'dt' in kwargs:

                # Ok, will convert the datetime to epoch and then create the point in the standard way
                t = s_from_dt(kwargs['dt'])

                # If we do not have a timezone, can we use the one from the dt used to initialize this TimePoint?
                try:
                    self._tz
                except AttributeError:
                    if kwargs['dt'].tzinfo:

                        #Do not set it if it is UTC, it is the default
                        if kwargs['dt'].tzinfo == UTC:
                            pass
                        else:
                            self._tz = kwargs['dt'].tzinfo
                            #raise NotImplementedError('Not yet tz from dt ("{}")'.format(kwargs['dt']))

            #else:
            #    raise Exception('Don\'t know how to handle all kwargs (got "{}")'.format(kwargs))

        # Cast or create in the standard way
        elif args:
            if len(args) > 1:
                raise Exception('Don\'t know how to handle all args (got "{}")'.format(args))

            if isinstance(args[0], TimePoint):
                t   = args[0].t
                self._tz = args[0].tz
            else:
                t = args[0]

        # Call parent init
        super(TimePoint, self).__init__(t)

    @property
    def t(self):
        """The timestamp as epoch, with decimals for sub-second precision."""
        return self.coordinates[0]

    def __gt__(self, other):
        if self.t > other.t:
            return True
        else:
            return False

    @property
    def tz(self):
        """The timezone."""
        try:
            return self._tz
        except AttributeError:
            return UTC

    def change_tz(self, tz):
        """Change the time zone of the point, in-place."""
        self._tz = timezonize(tz)

    @property
    def dt(self):
        """The timestamp as datetime object."""
        return dt_from_s(self.t, tz=self.tz)

    def __repr__(self):
        return 'Time point @ {} ({})'.format(self.t, self.dt)


class DataPoint(Point):
    """A point that carries some data. Data is attached using the respective data arguments.

       Args:
           *args (list): the coordinates.
           data: the data.
           data_indexes(dict): data indexes.
           data_loss(float): the data loss index, if any.
    """

    def __init__(self, *args, **kwargs):

        # Data
        try:
            self._data = kwargs.pop('data')
        except KeyError:
            raise Exception('A DataPoint requires a special "data" argument (got only "{}")'.format(kwargs))

        # Data indexes
        try:
            data_indexes = kwargs.pop('data_indexes')
            if data_indexes is None:
                data_indexes = {}
            else:
                if not isinstance(data_indexes, dict):
                    raise ValueError('Got type "{}" for data_indexes, was expecitng a dict'.format(data_indexes.__class__.__name__))
        except KeyError:
            data_indexes = {}
        #else:
        #    if None in data_indexes.values():
        #        raise ValueError('Cannot have an index set to None: do not set it at all ({})'.format(data_indexes))

        # Special data loss index
        try:
            data_loss = kwargs.pop('data_loss')
        except KeyError:
            pass
        else:
            data_indexes['data_loss'] = data_loss

        # Set data indexes
        self._data_indexes = data_indexes

        # Call parent init
        super(DataPoint, self).__init__(*args, **kwargs)

    def __repr__(self):
        if self.data_loss is not None:
            return 'Point @ {} with data "{}" and data_loss="{}"'.format(self.__coordinates_repr__, self.data, self.data_loss)
        else:
            return 'Point @ {} with data "{}"'.format(self.__coordinates_repr__, self.data)

    def __eq__(self, other):
        try:
            if self._data != other._data:
                return False
        except AttributeError:
            pass
        return super(DataPoint, self).__eq__(other)

    @property
    def data(self):
        """The data."""
        # Data is set like this as it cannot be set if not in the init (read: changed after created)
        # to prevent this to happend when the point is in a series where they are all supposed
        # to carry the same data type and with the same number of elements. TODO: check!
        return self._data

    @property
    def data_indexes(self):
        """The data indexes."""
        return self._data_indexes

    @data_indexes.setter
    def data_indexes(self, value):
        self._data_indexes = value

    @property
    def data_loss(self):
        """The data loss index, if any. Usually computed out from resampling transformations."""
        try:
            return self.data_indexes['data_loss']
        except KeyError:
            #raise AttributeError('No data loss index set for this point')
            return None

    def data_labels(self):
        """Returns the data labels. If data is a dictionary, then these are the dictionary keys, if data is
        list-like, then these are the list indexes (as strings). Other formats are not supported."""
        try:
            return sorted(list(self.data.keys()))
        except AttributeError:
            return [str(i) for i in range(len(self.data))]

    def _data_by_label(self, label):
        try:
            return self.data[label]
        except TypeError:
            try:
                return self.data[int(label)]
            except TypeError:
                raise TypeError('Cannot get an element by label on non-indexed data')


class DataTimePoint(DataPoint, TimePoint):
    """A point that carries some data in the time dimension. Can be initialized using the special `t` and `dt` arguments,
       for epoch seconds and datetime objects, respectively. Data is attached using the respective data arguments.

       Args:
           t(float): epoch timestamp, decimals for sub-second precision.
           dt(datetime): a datetime object timestamp.
           data: the data.
           data_indexes(dict): data indexes.
           data_loss(float): the data loss index, if any.
    """

    def __repr__(self):
        if self.data_loss is not None:
            return 'Time point @ {} ({}) with data "{}" and data_loss="{}"'.format(self.t, self.dt, self.data, self.data_loss)
        else:
            return 'Time point @ {} ({}) with data "{}"'.format(self.t, self.dt, self.data)



#======================
#  Slots
#======================

class Slot():
    """A slot. Can be initialized with start and end or start and unit.

       Args:
           start(Point): the slot starting point.
           end(Point): the slot ending point.
           unit(Unit): the slot unit.
    """
    __POINT_TYPE__ = Point
    __UNIT_TYPE__ = Unit

    def __init__(self, start, end=None, unit=None):

        if not isinstance(start, self.__POINT_TYPE__):
            raise TypeError('{} start must be a {} object (got "{}")'.format(self.__class__.__name__, self.__POINT_TYPE__.__name__, start.__class__.__name__))

        if end and not isinstance(end, self.__POINT_TYPE__):
            raise TypeError('{} end must be a {} object (got "{}")'.format(self.__class__.__name__, self.__POINT_TYPE__.__name__, end.__class__.__name__))

        if unit and not isinstance(unit, self.__UNIT_TYPE__):
            raise TypeError('{} unit must be a {} object (got "{}")'.format(self.__class__.__name__, self.__UNIT_TYPE__.__name__, unit.__class__.__name__))

        if end is None and unit is not None:
            if len(start.coordinates)>1:
                raise Exception('Sorry, setting a start and a unit only works in unidimensional spaces')
            end = start + unit
        else:
            if len(start.coordinates) != len(end.coordinates):
                raise ValueError('{} start and end dimensions must be the same (got "{}" vs "{}")'.format(self.__class__.__name__,start.coordinates, end.coordinates))
            if start == end:
                raise ValueError('{} start and end must not be the same (got start="{}", end="{}")'.format(self.__class__.__name__, start,end))

        self.start = start
        self.end   = end

        if unit is not None:
            self._unit = unit

    def __repr__(self):
        return 'Slot @ [{},{}]'.format(self.start.__coordinates_repr__, self.end.__coordinates_repr__)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if not self.start == other.start:
            return False
        if not self.end == other.end:
            return False
        return True

    def __succedes__(self, other):
        if other.end != self.start:
            return False
        else:
            return True

    @classmethod
    def _compute_length(cls, start, end):
        values = []
        for i in range(len(start.coordinates)):
            values.append(end.coordinates[i] - start.coordinates[i])

        return sum(values)/len(values)

    @property
    def length(self):
        """The slot length."""
        try:
            self._length
        except AttributeError:
            self._length = self._compute_length(self.start, self.end)
        return self._length

    @property
    def unit(self):
        """The slot unit."""
        try:
            return self._unit
        except AttributeError:
            if len(self.start.coordinates) == 1:
                return self.__UNIT_TYPE__(self.end.coordinates[0] - self.start.coordinates[0])
            else:
                values = []
                for i in range(len(self.start.coordinates)):
                    values.append(self.end.coordinates[i] - self.start.coordinates[i])

                return self.__UNIT_TYPE__(values)


class TimeSlot(Slot):
    """A slot in the time dimension. Can be initialized with start and end
       or start and unit.

       Args:
           start(TimePoint): the slot starting time point.
           end(TimePoint): the slot ending time point.
           unit(TimeUnit): the slot time unit."""

    __POINT_TYPE__ = TimePoint
    __UNIT_TYPE__ = TimeUnit

    def __init__(self, start=None, end=None, unit=None, **kwargs):

        # Internal-use init
        t = kwargs.get('t', None)
        dt = kwargs.get('dt', None)
        tz = kwargs.get('tz', None)

        # Handle t and dt shortcuts
        if t is not None:
            start=TimePoint(t=t, tz=tz)
        if dt is not None:
            start=TimePoint(dt=dt, tz=tz)

        # Extra time zone checks
        if start and end:
            if start.tz != end.tz:
                raise ValueError('{} start and end must have the same timezone (got start.tz="{}", end.tz="{}")'.format(self.__class__.__name__, start.tz, end.tz))

        # Call parent init
        super(TimeSlot, self).__init__(start=start, end=end, unit=unit)

        # If we did not have the end, set its timezone now:
        #if end is None:
        #    self.end.change_tz(self.start.tz)

        # Store time zone
        self.tz = start.tz


    # Overwrite parent succedes, this has better performance as it checks for only one dimension
    def __succedes__(self, other):
        if other.end.t != self.start.t:
            # Take into account floating point rounding errors
            if _is_close(other.end.t, self.start.t):
                return True
            return False
        else:
            return True

    def change_tz(self, tz):
        """Change the time zone of the slot, in-place."""
        self.start.change_tz(tz)
        self.end.change_tz(tz)
        self.tz = self.start.tz

    @property
    def unit(self):
        """The slot time unit"""
        try:
            return self._unit
        except AttributeError:
            # Use the string representation to handle floating point seconds
            self._unit = TimeUnit(str(self.end.t-self.start.t)+'s')
            return self._unit

    @property
    def t(self):
        """The slot epoch timestamp, intended as the starting point one."""
        return self.start.t

    @property
    def dt(self):
        """The slot datetime timestamp, intended as the starting point one."""
        return self.start.dt

    def __repr__(self):
        return 'Time slot @ [{},{}] ([{},{}])'.format(self.start.t, self.end.t, self.start.dt, self.end.dt)


class DataSlot(Slot):
    """A slot that carries some data. Can be initialized with start and end
       or start and unit, plus the data arguments.

       Args:
           start(Point): the slot starting point.
           end(Point): the slot ending point.
           unit(Unit): the slot unit.
           data: the data.
           data_indexes(dict): data indexes.
           data_loss(float): the data loss index, if any.
    """

    def __init__(self, *args, **kwargs):

        # Data
        try:
            self._data = kwargs.pop('data')
        except KeyError:
            raise Exception('A DataSlot requires a special "data" argument (got only "{}")'.format(kwargs))

        # Data indexes
        try:
            data_indexes = kwargs.pop('data_indexes')
            if data_indexes is None:
                data_indexes = {}
            else:
                if not isinstance(data_indexes, dict):
                    raise ValueError('Got type "{}" for data_indexes, was expecitng a dict'.format(data_indexes.__class__.__name__))
        except KeyError:
            data_indexes = {}
        #else:
        #    if None in data_indexes.values():
        #        raise ValueError('Cannot have an index set to None: do not set it at all ({})'.format(data_indexes))

        # Special data loss index
        try:
            data_loss = kwargs.pop('data_loss')
        except KeyError:
            pass
        else:
            data_indexes['data_loss'] = data_loss

        # Set data indexes
        self._data_indexes = data_indexes

        # Call parent init
        super(DataSlot, self).__init__(*args, **kwargs)

    def __repr__(self):
        if self.data_loss is not None:
            return 'Slot @ [{},{}] with data "{}" and data_loss="{}"'.format(self.start.__coordinates_repr__, self.end.__coordinates_repr__, self.data, self.data_loss)
        else:
            return 'Slot @ [{},{}] with data "{}"'.format(self.start.__coordinates_repr__, self.end.__coordinates_repr__, self.data)

    def __eq__(self, other):
        if self._data != other._data:
            return False
        return super(DataSlot, self).__eq__(other)

    @property
    def data(self):
        """The data."""
        # Data is set like this as it cannot be set if not in the init (read: changed after created)
        # to prevent this to happened when the point is in a series where they are all supposed
        # to carry the same data type and with the same number of elements. TODO: check me!
        return self._data

    @property
    def data_indexes(self):
        """The data indexes."""
        return self._data_indexes

    @data_indexes.setter
    def data_indexes(self, value):
        self._data_indexes = value

    @property
    def data_loss(self):
        """The data loss index, if any. Usually computed out from resampling transformations."""
        try:
            return self.data_indexes['data_loss']
        except KeyError:
            #raise AttributeError('No data loss index set for this point')
            return None

    def data_labels(self):
        """Returns the data labels. If data is a dictionary, then these are the dictionary keys, if data is
        list-like, then these are the list indexes (as strings). Other formats are not supported."""
        try:
            return sorted(list(self.data.keys()))
        except AttributeError:
            return [str(i) for i in range(len(self.data))]

    def _data_by_label(self, label):
        try:
            return self.data[label]
        except TypeError:
            try:
                return self.data[int(label)]
            except TypeError:
                raise TypeError('Cannot get an element by label on non-indexed data')


class DataTimeSlot(DataSlot, TimeSlot):
    """A slot that carries some data in the time dimension. Can be initialized
       with start and end or start and unit, plus the data arguments.

       Args:
           start(TimePoint): the slot starting time point.
           end(TimePoint): the slot ending time point.
           unit(TimeUnit): the slot time unit.
           data: the data.
           data_indexes(dict): data indexes.
           data_loss(float): the data loss index, if any.
    """

    def __repr__(self):
        if self.data_loss is not None:
            return 'Time slot @ [{},{}] ([{},{}]) with data={} and data_loss={}'.format(self.start.t, self.end.t, self.start.dt, self.end.dt, self.data, self.data_loss)
        else:
            return 'Time slot @ [{},{}] ([{},{}]) with data={}'.format(self.start.t, self.end.t, self.start.dt, self.end.dt, self.data)


#==============================
#  Series
#==============================

class Series(list):
    """A list of items coming one after another, where every item
       is guaranteed to be of the same type and in order or succession.

       The square brackets notation can be used for accessing series items, slicing the series
       or to filter it on a specific data label, if its elements support it:

           * ``series[3]`` will access the item in position #3;

           * ``series[3:5]`` will slice the series from item in position #3 to item in position #5 (excluded);

           * ``series['temperature']`` will filter the series keeping only the temperature data, assuming that
             in the original series there were also other data labels (e.g. humidity).

       Args:
           *args: the series items.
    """

    # 1,2,3 are in order. 1,3,8 are in order. 1,8,3 re not in order.
    # 5,6,7 are integer succession. 5.3, 5.4, 5.5 are too in a succesisons. 5,6,8 are not in a succession.
    # a group or a number of related or similar things, events, etc., arranged or occurring in temporal, spatial, or other order or succession; sequence.

    def __init__(self, *args, **kwargs):

        if kwargs:
            raise ValueError('Got an unknown argument "{}"'.format(list(kwargs.keys())[0]))

        for arg in args:
            self.append(arg)

    @property
    def item_type(self):
        """The type of the items of the series."""
        if not self:
            return None
        else:
            return self[0].__class__

    def append(self, item):
        """Append an item to the series. Accepts only items of the same
        type of the items already present in the series (unless empty)"""

        # TODO: move to use the insert?

        # Check type
        if self.item_type:
            if not isinstance(item, self.item_type):
                raise TypeError('Got incompatible type "{}", can only accept "{}"'.format(item.__class__.__name__, self.item_type.__name__))
        else:
            # First item appended to this series: check for compatible data type if any
            try:
                item.data
            except AttributeError:
                pass
            else:
                if _is_index_based(item.data):
                    if not _has_numerical_values(item.data):
                        logger.warning('You are using a data type that has no (or not only) numerical values. Support will be very limited.')
                elif _is_key_based(item.data):
                    if not _has_numerical_values(item.data):
                        logger.warning('You are using a data type that has no (or not only) numerical values. Support will be very limited.')
                else:
                    logger.warning('You are using a data type that is neither index-based nor in key-value format. Support will be very limited.')

        # Check order or succession if not empty
        if self:
            try:
                if not item.__succedes__(self[-1]):
                    raise ValueError('Not in succession ("{}" does not succeedes "{}")'.format(item,self[-1])) from None
            except IndexError:
                raise
            except AttributeError:
                try:
                    if not item > self[-1]:
                        raise ValueError('Not in order ("{}" does not follow "{}")'.format(item,self[-1])) from None
                except TypeError:
                    raise TypeError('Object of class "{}" does not implement a "__gt__" or a "__succedes__" method, cannot append it to a Series (which is ordered)'.format(item.__class__.__name__)) from None

        # Check data type and set if not done already
        if self and self._item_data_reference:
            if not type(self._item_data_reference) == type(item.data):
                raise TypeError('Got different data: {} vs {}'.format(self._item_data_reference.__class__.__name__, item.data.__class__.__name__))
            if isinstance(self._item_data_reference, list):
                if len(self._item_data_reference) != len(item.data):
                    raise ValueError('Got different data lengths: {} vs {}'.format(len(self._item_data_reference), len(item.data)))
            if isinstance(self._item_data_reference, dict):
                if set(self._item_data_reference.keys()) != set(item.data.keys()):
                    raise ValueError('Got different data labels: {} vs {}'.format(self._item_data_reference.keys(), item.data.keys()))
        else:
            try:
                self._item_data_reference = item.data
            except AttributeError:
                self._item_data_reference = None

        # Append
        super(Series, self).append(item)

    def __sum__(self, other):
        raise NotImplementedError

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return '{} of #{} elements'.format(self.__class__.__name__, len(self))

    def __str__(self):
        return self.__repr__()

    def _all_data_indexes(self):
        """Return all the data_indexes of the series, to be intended as custom
        defined indicators (i.e. data_loss, anomaly_index, etc.)."""

        # TODO: move this to the Data*Series...?
        data_index_names = []
        for item in self:
            for index_name in item.data_indexes:
                if index_name not in data_index_names:
                    data_index_names.append(index_name)

        # Reorder according to legacy indexes behaviour
        ordered_data_index_names = []
        legacy_indexes = ['data_reconstructed', 'data_loss', 'anomaly', 'forecast']
        for legacy_index in legacy_indexes:
            if legacy_index in data_index_names:
                ordered_data_index_names.append(legacy_index)
                data_index_names.remove(legacy_index)

        # Merge and sort
        ordered_data_index_names += sorted(data_index_names)

        return ordered_data_index_names

    # Inherited methods to be edited
    def insert(self, i, x):
        """Insert an item at a given position. The first argument is the index of the element
        before which to insert, so series.insert(0, x) inserts at the front of the series, and
        series.insert(len(series), x) is equivalent to append(x). Order or succession are enforced."""

        if len(self) > 0:

            # Check valid type
            if not isinstance(x, self.item_type):
                raise TypeError('Got incompatible type "{}", can only accept "{}"'.format(x.__class__.__name__, self.item_type.__name__))

            # Check ordering/succession
            if i == 0:
                try:
                    if not self[0].__succedes__(x):
                        raise ValueError('Cannot insert element "{}" in position "{}" as it would break the succession'.format(x,i))
                except AttributeError:
                    if not self[0] > x:
                        raise ValueError('Cannot insert element "{}" in position "{}" as not in order'.format(x,i)) from None

            elif i == len(self):
                try:
                    if not x.__succedes__(self[-1]):
                        raise ValueError('Cannot insert element "{}" in position "{}" as it would break the succession'.format(x,i))
                except AttributeError:
                    if not x > self[-1]:
                        raise ValueError('Cannot insert element "{}" in position "{}" as not in order'.format(x,i)) from None

            else:
                try:
                    self[0].__succedes__
                except AttributeError:
                    if not x > self[i-1]:
                        raise ValueError('Cannot insert element "{}" in position "{}" with element "{}" as not in order'.format(x,i, self[i])) from None
                else:
                    raise IndexError('Cannot insert an item in the middle of a series whose items follow a succession'.format(i))

            super(Series, self).insert(i,x)

        else:
            self.append(x)

    def remove(self, x):
        """Remove the first item from the list whose value is equal to x. It raises a `ValueError` if there is no such item
        and a `NotImplementedError` is the series items are in a succession as it would breake it.
        """
        try:
            self[0].__succedes__
            raise NotImplementedError('Remove is not implemented for series whose items are in a succession')
        except IndexError:
            pass
        except AttributeError:
            pass

        super(Series, self).remove(x)

    def pop(self, i=None):
        """
        Remove the item at the given position in the list, and return it. If no index is
        specified, removes and returns the last item in the series. If items are in a
        succession and the index is set, a `NotImplementerError` is raised.
        """
        if len(self) == 0:
            raise IndexError('Cannot pop from an empty series')
        if i is not None:
            if i not in [0, len(self)]:
                try:
                    self[0].__succedes__
                except AttributeError:
                    pass
                else:
                    raise IndexError('Cannot pop an item in the middle of a series whose items follow a succession')
            return super(Series, self).pop(i)
        else:
            return super(Series, self).pop()

    def clear(self):
        """Remove all items from the series."""
        super(Series, self).clear()

    def index(self, x, start=None, end=None):
        """
        Return zero-based index in the series of the item whose value is equal to x.
        Raises a ValueError if there is no such item.

        The optional arguments start and end are interpreted as in the slice notation and are
        used to limit the search to a particular subsequence of the list. The returned index is
        computed relative to the beginning of the full sequence rather than the start argument.
        """
        super(Series, self).index(x, start, end)

    def copy(self):
        """Return a shallow copy of the series."""
        return super(Series, self).copy()

    def duplicate(self):
        """ Return a deep copy of the series."""
        return deepcopy(self)

    def _item_by_i(self, i):
        return super(Series, self).__getitem__(i)

    def _slice_by_i(self, **args):
        return super(Series, self).__getitem__(**args)

    # Inherited methods to be disabled
    def extend(self):
        """Disabled (use the `merge` instead)."""
        raise NotImplementedError('Use the "merge()" instead')

    def count(self, x):
        """Disabled (there is only one item instance by design)."""
        raise NotImplementedError('There is only one item by design')

    def sort(self, key=None, reverse=False):
        """Disabled (sorting is already guaranteed)."""
        raise NotImplementedError('Sorting is already guaranteed')

    def reverse(self):
        """Disabled (reversing is not compatible with an ordering)."""
        raise NotImplementedError('Reversing is not compatible with an ordering')


    #=========================
    # Square brackets notation
    #=========================

    def __getitem__(self, key):
        if isinstance(key, slice):
            indices = range(*key.indices(len(self)))
            # Prepare the new series and return it
            series = self.__class__()
            for i in indices:
                series.append(super(Series, self).__getitem__(i))
            return series
        elif isinstance(key, str):
            # Try filtering on this data label only
            return self.filter(key)
        else:
            # TOOD: this will not work for SeriesView if ever implemented
            return super(Series, self).__getitem__(key)


    #=========================
    #  Data-related
    #=========================

    def data_labels(self):
        """Returns the labels of the data carried by the series items. If data is a dictionary, then
        these are the dictionary keys, if data  is list-like, then these are the list indexes
        (as strings). Other formats are not supported."""
        if len(self) > 0 and not self._item_data_reference:
            raise TypeError('Series items have no data, cannot get data labels')
        if len(self) == 0:
            return None
        else:
            return self[0].data_labels()

    def rename_data_label(self, old_data_label, new_data_label):
        """Rename a data label, in-place."""
        if len(self) > 0 and not self._item_data_reference:
            raise TypeError('Series items have no data, cannot rename a label')
        for item in self:
            # TODO: move to the DataPoint/DataSlot?
            item.data[new_data_label] = item.data.pop(old_data_label)

    def remove_data_label(self, data_label):
        """Remove a data label, in-place."""
        if len(self) > 0 and not self._item_data_reference:
            raise TypeError('Series items have no data, cannot rename a label')
        for item in self:
            # TODO: move to the DataPoint/DataSlot?
            item.data.pop(data_label, None)

    def remove_data_index(self, data_index):
        """Remove a data index, in-place."""
        if len(self) > 0 and not self._item_data_reference:
            raise TypeError('Series items have no data, cannot rename data indexes')
        for item in self:
            item.data_indexes.pop(data_index, None)

    def remove_data_loss(self):
        """Remove the ``data_loss`` index, in-place."""
        if len(self) > 0 and not self._item_data_reference:
            raise TypeError('Series items have no data, cannot remove the data loss')
        for item in self:
            item.data_indexes.pop('data_loss', None)




    #=========================
    #  Operations
    #=========================

    def min(self, data_label=None):
        """Get the minimum data value(s) of a series. Supports an optional ``data_label`` argument.
        A series of DataPoints or DataSlots is required."""
        from .operations import min as min_operation
        return min_operation(self, data_label=data_label)

    def max(self, data_label=None):
        """Get the maximum data value(s) of a series. Supports an optional ``data_label`` argument.
        A series of DataPoints or DataSlots is required."""
        from .operations import max as max_operation
        return max_operation(self, data_label=data_label)

    def avg(self, data_label=None):
        """Get the average data value(s) of a series. Supports an optional ``data_label`` argument.
        A series of DataPoints or DataSlots is required."""
        from .operations import avg as avg_operation
        return avg_operation(self, data_label=data_label)

    def sum(self, data_label=None):
        """Sum every data value(s) of a series. Supports an optional ``data_label`` argument.
        A series of DataPoints or DataSlots is required."""
        from .operations import sum as sum_operation
        return sum_operation(self, data_label=data_label)

    def derivative(self, inplace=False, normalize=True, diffs=False):
        """Compute the derivative of the series. Extra parameters: ``inplace`` (defaulted to false),
        ``normalize`` (defaulted to true) and ``diffs`` (defaulted to false) to compute differences
        instead of the derivative. A series of DataTimePoints or DataTimeSlots is required."""
        from .operations import derivative as derivative_operation
        return derivative_operation(self, inplace=inplace, normalize=normalize, diffs=diffs)

    def integral(self, inplace=False, normalize=True, c=0, offset=0):
        """Compute the integral of the series. Extra parameters: ``inplace`` (defaulted to false),
        ``normalize`` (defaulted to true), ``c`` (defaulted to zero) for the integration constant
        and ``offset`` (defaulted to zero) to start the integration from an offset.
        A series of DataTimePoints or DataTimeSlots is required."""
        from .operations import integral as integral_operation
        return integral_operation(self, inplace=inplace, normalize=normalize, c=c, offset=offset)

    def diff(self, inplace=False):
        """Compute the incremental differences. Reduces the series length by one (the first element).
        Extra parameters: ``inplace`` (defaulted to false). A series of DataPoints or DataSlots is required."""
        from .operations import diff as diff_operation
        return diff_operation(self, inplace=inplace)

    def csum(self, inplace=False, offset=None):
        """Compute the incremental sum. Extra parameters: ``inplace`` (defaulted to false),
        ``offset`` (defaulted to zero) to set the starting value where to apply the sums on.
        A series of DataPoints or DataSlots is required."""
        from .operations import csum as csum_operation
        return csum_operation(self, inplace=inplace, offset=offset)

    def normalize(self, range=[0,1], inplace=False, source_range=None):
        """Normalize the series data values. Extra parameters: ``range`` (defaulted to [0,1]) to set the normalization
        target range, ``source_range`` to define a custom source range to normalize with respect to, and ``inplace``
        (defaulted to false). A series of DataPoints or DataSlots is required."""
        from .operations import normalize as normalize_operation
        return normalize_operation(self, inplace=inplace, range=range, source_range=source_range)

    def rescale(self, value, inplace=False):
        """Rescale the series data values by a ``value``. This can be either a single number or a dictionary
        where to set rescaling factors on a per-data label basis. Extra parameters: ``inplace`` (defaulted to false)
        A series of DataPoints or DataSlots is required."""
        from .operations import rescale as rescale_operation
        return rescale_operation(self, value=value, inplace=inplace)

    def offset(self, value, inplace=False):
        """Offset the series data values by a ``value``. This can be either a single number or a dictionary
        where to set offsetting factors on a per-data label basis. Extra parameters: ``inplace`` (defaulted to false).
        A series of DataPoints or DataSlots is required."""
        from .operations import offset as offset_operation
        return offset_operation(self, value=value, inplace=inplace)

    def mavg(self,  window, inplace=False):
        """Compute the moving average. Reduces the series length by n (the window size). Extra
        parameters: ``inplace`` (defaulted to false) and ``window``, a required parameter, for
        the size of the moving average window. A series of DataPoints or DataSlots is required."""
        from .operations import mavg as mavg_operation
        return mavg_operation(self, window=window, inplace=inplace)

    def merge(self, series):
        """Merge the series with one or more other series."""
        from .operations import merge as merge_operation
        return merge_operation(self, series)

    def get(self, at_i=None):
        from .operations import get as get_operation
        return get_operation(self, at_i)

    def filter(self, *data_labels):
        """Filter a series by data label(s). A series of DataPoints or DataSlots is required."""
        from .operations import filter as filter_operation
        return filter_operation(self, *data_labels)

    def slice(self, from_i=None, to_i=None):
        """Slice a series. A series of DataPoints or DataSlots is required."""
        from .operations import slice as slice_operation
        return slice_operation(self, from_i=from_i, to_i=to_i)

    def select(self, query):
        """Select one or more items of the series given an SQL-like query. A series of DataPoints or DataSlots is required."""
        from .operations import select as select_operation
        return select_operation(self, query=query)


    #=========================
    #  Transformations
    #=========================

    def aggregate(self, unit, *args, **kwargs):
        """Aggregate the series in slots of length set by the ``unit`` parameter. A series of DataPoints or DataSlots is required."""
        from .transformations import Aggregator
        aggregator = Aggregator(unit, *args, **kwargs)
        return aggregator.process(self)

    def resample(self, unit, *args, **kwargs):
        """Resample the series using a sampling interval of length set by the ``unit`` parameter. A series of DataPoints or DataSlots is required."""
        from .transformations import Resampler
        resampler = Resampler(unit, *args, **kwargs)
        return resampler.process(self)


    #=========================
    # Inspection utilities
    #=========================

    def _summary(self, limit=10):

        string = str(self)+':\n\n'
        string+='['

        if not limit or limit > len(self):

            for i, item in enumerate(self):
                if limit and i >= limit:
                    break
                else:
                    if i==0:
                        string+=str(item)+',\n'
                    elif i==len(self)-1:
                        string+=' '+str(item)
                    else:
                        string+=' '+str(item)+',\n'
        else:
            if limit==1:
                head_n=1
                tail_n=0
            else:
                if limit % 2 == 0:
                    head_n = int(limit/2)
                else:
                    head_n = int(limit/2)+1
                tail_n = int(limit/2)

            for i, item in enumerate(self.head(head_n)):
                if i==0:
                    string+=str(item)+',\n'
                else:
                    string+=' '+str(item)+',\n'
            if limit < len(self):
                string+=' ...\n'
            if tail_n != 0:
                for i, item in enumerate(self.tail(tail_n)):
                    if i==tail_n-1:
                        string+=' '+str(item)
                    else:
                        string+=' '+str(item)+',\n'

        string+=']'
        return string

    def summary(self, limit=10, newlines=False):
        """Return a string summary of the series and its elements, limited to 10 items by default.

            Args:
                limit: the limit of elements to print, by default 10.
                newlines: if to include the newline characters or not.
        """
        if newlines:
            return self._summary(limit=limit)
        else:
            return self._summary(limit=limit).replace('\n', ' ').replace('...', '...,')

    def inspect(self, limit=10):
        """Print a summary of the series and its elements, limited to 10 items by default.

            Args:
                limit: the limit of elements to print, by default 10.
        """
        print(self._summary(limit=limit))

    def contents(self):
        """Get all the items of the series as a list."""
        return list(self)

    def head(self, n=5):
        """Get the first n items of the series as a list, 5 by default.

            Args:
                n: the number of first elements to return .
        """
        return list(self[0:n])

    def tail(self, n=5):
        """Get the last n items of the series as a list, 5 by default.

            Args:
                n: the number of last elements to return .
        """
        return list(self[-n:])



#==============================
#  Time Series
#==============================

class TimeSeries(Series):
    """A list of items coming one after another over time, where every item
       is guaranteed to be of the same type and in order or succession.

       Time series accept only items of type :obj:`DataTimePoint` and :obj:`DataTimeSlot`
       (or :obj:`TimePoint` and :obj:`TimeSlot` which are useful in some circumstances),
       but can be created using some shortcuts, for example:

           * providing a Pandas Dataframe with a time-based index;

           * providing a list of dictionaries in the following forms, plus an optional ``slot_unit`` argument
             if creating a slot series (e.g. ``slot_unit='1D'``):

               * ``{60: 4, 120: 6, ... }``
               * ``{dt(1970,1,1): 4, dt(1970,1,2): 6, ... }``

           * providing a string with a path to a CSV file, which will be read and parsed using a :obj:`timeseria.storages.CSVStorage`
             storage object (in this case all the key-value arguments will be forwarded to the storage).

       The square brackets notation can be used for accessing series items, slicing the series
       or to filter it on a specific data label (if the elements support it), as outlined below.

           * ``series[3]`` will access the item in position #3;

           * ``series[1446073200.7]`` will access the item for the epoch timestamp corresponding to the (floating point)
             number provided in the square brackets;

           * ``series[dt(2015,10,25,6,19,0)]`` will access the item for the corresponding datetime timestamp.

       The same three options can be used for slicing, and filtering a series on a data label can also be achieved using
       the square bracket notation, by providing the data label on which to filter the series: ``series['temperature']`` will filter
       the time series keeping only temperature data, assuming that in the original series there were also other data labels (e.g. humidity).

       For more options for accessing and selecting series items and for slicing or filtering series, see the corresponding
       methods: :func:`select()`, :func:`slice()` and :func:`filter()`.

       Args:
           *args: the time series items, or the right object for an alternative init method as described above.
    """

    def __repr__(self):
        if not self:
            return 'Empty time series'
        else:
            if issubclass(self.item_type, TimePoint):
                return 'Time series of #{} points at {}, from point @ {} ({}) to point @ {} ({})'.format(len(self), self._resolution_string, self[0].t, self[0].dt, self[-1].t, self[-1].dt)
            elif issubclass(self.item_type, TimeSlot):
                return 'Time series of #{} slots of {}, from slot starting @ {} ({}) to slot starting @ {} ({})'.format(len(self), self.resolution, self[0].start.t, self[0].start.dt, self[-1].start.t, self[-1].start.dt)
            else:
                raise ConsistencyException('Got no TimePoints nor TimeSlots in a Time Series, this is a consistency error (got {})'.format(self.item_type.__name__))


    #=========================
    #  Init
    #=========================

    def __init__(self, *args, **kwargs):


        # Handle timezone
        tz = kwargs.pop('tz', None)
        if tz:
            self._tz = timezonize(tz)

        # Call parent init
        super(TimeSeries, self).__init__(*args, **kwargs)


    #=========================
    #  Append
    #=========================

    def append(self, item):
        """Append an item to the time series. Accepts only items of type :obj:`DataTimePoint` and :obj:`DataTimeSlot`
        (or :obj:`TimePoint` and :obj:`TimeSlot`, which are useful in some particular circumstances) and in any case
        of the same type of the items already present in the time series, unless empty."""

        if isinstance(item, TimePoint):

            try:
                self.prev_t
            except AttributeError:
                pass
            else:
                # Check time ordering and handle the resolution.

                # The following if is to support the deepcopy, otherwise the original prev_t will be used
                # TODO: maybe move the above to a "hasattr" plus an "and" instead of this logic?
                if len(self)>0:

                    # logger.debug('Checking time ordering for t="%s" (prev_t="%s")', item.t, self.prev_t)
                    if item.t < self.prev_t:
                        raise ValueError('Time t="{}" is out of order (prev t="{}")'.format(item.t, self.prev_t))

                    if item.t == self.prev_t:
                        raise ValueError('Time t="{}" is a duplicate'.format(item.t))

                    try:
                        self._resolution

                    except AttributeError:

                        # Set the resolution as seconds for a bit of performance
                        self._resolution_as_seconds = item.t - self.prev_t

                        # Set the resolution as time unit (up to the microsecond)
                        time_diff = item.t - self.prev_t
                        time_diff = int(time_diff*10000000)/10000000
                        self._resolution = TimeUnit(_to_time_unit_string(time_diff, friendlier=True))

                    else:
                        # If the resolution is constant (not variable), check that it still is
                        if self._resolution != 'variable':
                            if self._resolution_as_seconds != (item.t - self.prev_t):
                                # ...otherwise, mark it as variable
                                del self._resolution_as_seconds
                                self._resolution = 'variable'
            finally:
                # Delete the autodetected sampling interval cache if present
                try:
                    del self._autodetected_sampling_interval
                    del self._autodetected_sampling_interval_confidence
                except:
                    pass
                # And set the prev
                self.prev_t = item.t

        elif isinstance(item, TimeSlot):

            # Slots can belong to the same series if they are in succession (tested with the __succedes__ method)
            # and if they have the same unit, which we test here instead as the __succedes__ is more general.

            # Check the timezone (only for slots, points are not affected by timezones)
            if not self.tz:
                # If no timezone set, use the item one's
                self._tz = item.tz

            else:
                # Else, check for the same timezone
                if self._tz != item.tz:
                    raise ValueError('Cannot append slots on different timezones (I have "{}" and you tried to add "{}")'.format(self.tz, item.start.tz))

            try:
                if self._resolution != item.unit:
                    # Try for floating point precision errors
                    abort = False
                    try:
                        if not _is_close(self._resolution, item.unit):
                            abort = True
                    except (TypeError, ValueError, NotImplementedError):
                        abort = True
                    if abort:
                        raise ValueError('Cannot add slots with a different unit than the series resolution (I have "{}" and you tried to add "{}")'.format(self._resolution, item.unit))
            except AttributeError:
                self._resolution = item.unit

        else:
            raise TypeError('Adding data to a time series only accepts TimePoints o TimeSlots (got "{}")'.format(item.__class__.__name__))

        # Lastly, call the series append
        super(TimeSeries, self).append(item)

    def _item_by_t(self, t):
        # TODO: improve performance, bisection first, then use an index?
        for item in self:
            if item.t == t:
                return item
        raise ValueError('Cannot find any item for t={}'.format(t))


    #=========================
    # Square brackets notation
    #=========================

    def __getitem__(self, arg):
        if isinstance(arg, slice):
            if isinstance(arg.start, int) or isinstance(arg.stop, int):
                return self.slice(from_i=arg.start, to_i=arg.stop)
            elif isinstance(arg.start, float) or isinstance(arg.stop, float):
                if arg.start is not None and not (isinstance(arg.start, Time)):
                    logger.warning('Slicing in the square brackets notation with a float works (as epoch) but it can be ambiguous in the code')
                if arg.stop is not None and not (isinstance(arg.stop, Time)):
                    logger.warning('Slicing in the square brackets notation with a float works (as epoch) but it can be ambiguous in the code')
                return self.slice(from_t=arg.start, to_t=arg.stop)
            elif isinstance(arg.start, datetime) or isinstance(arg.stop, datetime):
                return self.slice(from_dt=arg.start, to_dt=arg.stop)
            else:
                raise ValueError('Don\'t know how to slice for data type "{}"'.format(arg.start.__class__.__name__))
        elif isinstance(arg, str):
            return self.filter(arg)
        else:
            if isinstance(arg, int):
                return self.get(at_i=arg)
            elif isinstance(arg, float):
                if not isinstance(arg, Time):
                    logger.warning('Getting items in the square brackets notation with a float works (as epoch) but it can be ambiguous in the code')
                return self.get(at_t=arg)
            elif isinstance(arg, datetime):
                return self.get(at_dt=arg)
            else:
                raise ValueError('Don\'t know how to slice for data type "{}"'.format(arg.start.__class__.__name__))
            return self.get(arg)

    #=========================
    #  Timezone-related
    #=========================

    @property
    def tz(self):
        """The timezone of the time series."""
        # Note: we compute the tz on the fly because for point time series we assume to use the tz
        # attribute way lass than the slot time series, where the tz is instead computed at append-time.
        try:
            return self._tz
        except AttributeError:
            # Detect timezone on the fly
            # TODO: this ensures that each point is on the same timezone. Do we want this?
            detected_tz = None
            for item in self:
                if not detected_tz:
                    detected_tz = item.tz
                else:
                    # Terrible but seems like no other way to compare pytz.tzfile.* classes
                    if str(item.tz) != str(detected_tz):
                        return UTC
            return detected_tz

    def change_tz(self, tz):
        """Change the time zone of the time series, in-place."""
        for time_point in self:
            time_point.change_tz(tz)
        self._tz = time_point.tz

    def as_tz(self, tz):
        """Get a copy of the time series on a new time zone."""
        new_series = self.duplicate()
        new_series.change_tz(tz)
        return new_series


    #=========================
    #  Resolution-related
    #=========================

    @property
    def _autodetected_sampling_interval(self):
        if not issubclass(self.item_type, TimePoint):
            raise NotImplementedError('Auto-detecting the sampling rate (and its confidence) is implemented only for point series')
        try:
            return self.__autodetected_sampling_interval
        except AttributeError:
            from .utils import detect_sampling_interval
            self.__autodetected_sampling_interval, self.__autodetected_sampling_interval_confidence = detect_sampling_interval(self, confidence=True)
            return self.__autodetected_sampling_interval

    @property
    def _autodetected_sampling_interval_confidence(self):
        if not issubclass(self.item_type, TimePoint):
            raise NotImplementedError('Auto-detecting the sampling rate (and its confidence) is implemented only for point series')
        try:
            return self.__autodetected_sampling_interval_confidence
        except AttributeError:
            from .utils import detect_sampling_interval
            self.__autodetected_sampling_interval, self.__autodetected_sampling_interval_confidence = detect_sampling_interval(self, confidence=True)
            return self.__autodetected_sampling_interval_confidence

    @property
    def resolution(self):
        """The (temporal) resolution of the time series.

        Returns a :obj:`timeseria.units.TimeUnit` object, unless:

            * the resolution is not defined (returns :obj:`None`), either because the time series is empty or because it is a point time series with only one point; or
            * the resolution is variable (returns the string ``variable``), only possible for point time series, if its points are not equally spaced, for example because of
              data losses or uneven observations.

        If the time series has a variable resolution, the `guess_resolution()` method can provide an estimate. If the time series is a slot time series, then the resolutions is just
        the unit of its slots.
        """
        try:
            return self._resolution
        except AttributeError:
            # Case of an empty or 1-point series
            return None

    def guess_resolution(self, confidence=False):
        """Guess the (temporal) resolution of the time series.

           Args:
               confidence (bool): if to return, together with the guessed resolution, also its confidence (in a 0-1 range).

        """
        if not issubclass(self.item_type, TimePoint):
            raise NotImplementedError('Guessing the resolution is implemented only for point series')
        if not self:
            raise ValueError('Cannot guess the resolution for an empty time series')
        if len(self) == 1:
            raise ValueError('Cannot guess the resolution for a time series with only one point')
        if self.resolution != 'variable':
            raise ValueError('The time series has a well defined resolution ({}), guessing it does not make sense'.format(self.resolution))
        else:
            try:
                self._guessed_resolution
            except AttributeError:
                self._guessed_resolution = TimeUnit(_to_time_unit_string(self._autodetected_sampling_interval, friendlier=True))
            finally:
                if confidence:
                    return {'value': self._guessed_resolution, 'confidence': self._autodetected_sampling_interval_confidence}
                else:
                    return self._guessed_resolution

    @property
    def _resolution_string(self):
        if self.resolution is None:
            resolution_string = 'undefined resolution'
        else:
            if self.resolution == 'variable':
                autodetected_sampling_interval_as_str = TimeUnit(_to_time_unit_string(self._autodetected_sampling_interval, friendlier=True))
                resolution_string = 'variable resolution (~{})'.format(autodetected_sampling_interval_as_str)
            else:
                resolution_string = '{} resolution'.format(self.resolution)
        return resolution_string


    #=========================
    #  Load/save
    #=========================

    @classmethod
    def load(cls, file_name):
        """Load a series from a (CSV) file, in Timeseria format."""

        metadata = None
        with open(file_name) as f:
            for line in f:
                if line.startswith('#'):
                    if line.startswith('# Extra parameters:'):
                        metadata = json.loads(line.replace('# Extra parameters:',''))
                else:
                    break
        if not metadata:
            raise ValueError('The file provided is not a Timeseria time series data file. Perhaps you wanted to use the from_csv() method?')

        force_points = True if metadata['type'] == 'points' else False
        force_slots = True if metadata['type'] == 'slots' else False
        force_tz = metadata['tz']
        if force_slots:
            force_slot_unit = metadata['resolution']
        else:
            force_slot_unit = None

        from .storages import CSVFileStorage
        storage = CSVFileStorage(file_name)
        loaded_series = storage.get(force_points=force_points, force_slots=force_slots, force_tz=force_tz, force_slot_unit=force_slot_unit)

        if loaded_series.__class__ == cls:
            return loaded_series
        else:
            # TODO: the following is a huge performance hit...
            series_items = loaded_series.contents()
            cls(*series_items)


    def save(self, file_name, overwrite=False, **kwargs):
        """Save the time series as a (CSV) file, in Timeseria format."""
        from .storages import CSVFileStorage
        storage = CSVFileStorage(file_name, **kwargs)
        storage.put(self, overwrite=overwrite)


    #=========================
    #  Conversions
    #=========================

    @classmethod
    def from_dict(cls, dictionary, slot_unit=None):
        """Create a time series from a dictionary."""

        if slot_unit and not isinstance(slot_unit, Unit):
                slot_unit = TimeUnit(slot_unit)
        series = cls()

        for key in dictionary:
            item_dt = None
            item_t = None
            if isinstance(key, datetime):
                item_dt = key
            elif isinstance(key, int) or isinstance(key,float):
                item_t = key
            elif isinstance(key, str):
                try:
                    item_t = float(key)
                except ValueError:
                    try:
                        item_t = s_from_dt(dt_from_str(key))
                    except Exception as e:
                        raise e from None
            else:
                raise ValueError('Cannot handle keys other than int, float or datetime (got "{}"'.format(key.__class__.__name__))
            data = dictionary[key]
            if not (isinstance(data, list) or isinstance(data, dict)):
                data = {'value': data}

            # Create the item
            if slot_unit:
                if item_dt is not None:
                    series.append(DataTimeSlot(dt=item_dt, unit=slot_unit, data=data))
                else:
                    series.append(DataTimeSlot(t=item_t, unit=slot_unit, data=data))
            else:
                if item_dt is not None:
                    series.append((DataTimePoint(dt=item_dt, data=data)))
                else:
                    series.append((DataTimePoint(t=item_t, data=data)))

        return series 

    def to_dict(self):
        """Convert a time series to a dictionary."""
        timeseries_as_dict = {}
        for item in self:
            timeseries_as_dict[item.dt] = item.data
        return timeseries_as_dict


    @classmethod
    def from_json(cls, string, slot_unit=None):
        """Create a time series from a JSON string."""
        json_data = json.loads(string)
        return cls.from_dict(json_data, slot_unit=slot_unit)


    def to_json(self):
        """Convert a time series to a JSON string."""
        timeseries_as_dict = {}
        for item in self:
            timeseries_as_dict[str_from_dt(item.dt)] = item.data
        return json.dumps(timeseries_as_dict)


    @classmethod
    def from_csv(cls, file_name, *args, **kwargs):
        """Create a a time series from a CSV file."""
        from .storages import CSVFileStorage
        storage = CSVFileStorage(file_name, *args, **kwargs)
        loaded_series = storage.get()

        if loaded_series.__class__ == cls:
            return loaded_series
        else:
            # TODO: the following is a huge performance hit...
            series_items = loaded_series.contents()
            cls(*series_items)

    def to_csv(self, file_name, overwrite=False, **kwargs):
        """Store the time series as a CSV file."""
        from .storages import CSVFileStorage
        storage = CSVFileStorage(file_name, **kwargs)
        storage.put(self, overwrite=overwrite)


    @classmethod
    def from_df(cls, df, item_type='auto'):
        """Create a time series from a Pandas data frame."""

        if item_type == 'auto':
            item_type = None

        # Infer if we have to create points or slots and their unit
        unit_str_pd=df.index.inferred_freq

        if not unit_str_pd:
            if not item_type:
                logger.info('Cannot infer the freqency of the dataframe, will just create points')
                item_type = DataTimePoint

        else:

            # Calendar
            unit_str_pd=unit_str_pd.replace('A', 'Y')    # Year (end) ?
            unit_str_pd=unit_str_pd.replace('Y', 'Y')    # Year (end) )
            unit_str_pd=unit_str_pd.replace('AS', 'Y')   # Year (start)
            unit_str_pd=unit_str_pd.replace('YS', 'Y')   # Year (start)
            unit_str_pd=unit_str_pd.replace('MS', 'M')   # Month (start)
            unit_str_pd=unit_str_pd.replace('M', 'M')    # Month (end)
            unit_str_pd=unit_str_pd.replace('D', 'D')    # Day

            # Physical
            unit_str_pd=unit_str_pd.replace('H', 'h')    # Hour
            unit_str_pd=unit_str_pd.replace('T', 'm')    # Minute
            unit_str_pd=unit_str_pd.replace('min', 'm')  # Minute
            unit_str_pd=unit_str_pd.replace('S', 's')    # Second

            if len(unit_str_pd) == 1:
                unit_str = '1'+unit_str_pd
            else:
                unit_str = unit_str_pd

            unit_type = ''.join([char for char in unit_str if not char.isdigit()])

            if unit_type in ['Y', 'M', 'D']:
                logger.info('Assuming slots with a slot time unit of "{}"'.format(unit_str))
                if not item_type:
                    item_type = DataTimeSlot
                else:
                    if item_type==DataTimePoint:
                        raise ValueError('Creating points with calendar time units is not supported.')
            else:
                if not item_type:
                    item_type = DataTimePoint

        # Now create the points or the slots
        if item_type==DataTimeSlot:

            # Init the unit
            unit = TimeUnit(unit_str)

            # Create data time points list
            datatimeslots = []

            # Get data frame labels
            labels = list(df.columns)

            naive_warned = False
            for row in df.iterrows():

                # Set the timestamp
                if not isinstance(row[0], datetime):
                    raise TypeError('A DataFrame with a DateTime index column is required')
                dt = row[0]

                # Prepare data
                data = {}
                for i,label in enumerate(labels):
                    data[label] = row[1][i]

                if dt.tzinfo is None:
                    if not naive_warned:
                        logger.warning('Got naive datetimes as dataframe index, assuming UTC.')
                        naive_warned = True
                    dt = UTC.localize(dt)

                datatimeslots.append(DataTimeSlot(dt=dt, unit=unit, data=data))

            # Set the list of data time points
            return cls(*datatimeslots)

        else:
            # Create data time points list
            datatimepoints = []

            # Get data frame labels
            labels = list(df.columns)

            naive_warned = False
            for row in df.iterrows():

                # Set the timestamp
                if not isinstance(row[0], datetime):
                    raise TypeError('A DataFrame with a DateTime index column is required')
                dt = row[0]

                # Prepare data
                data = {}
                for i,label in enumerate(labels):
                    data[label] = row[1][i]

                if dt.tzinfo is None:
                    if not naive_warned:
                        logger.warning('Got naive datetimes as dataframe index, assuming UTC.')
                        naive_warned = True
                    dt = UTC.localize(dt)

                datatimepoints.append(DataTimePoint(dt=dt, data=data))

            # Set the list of data time points
            return cls(*datatimepoints)

    def to_df(self):
        """Convert the time series as a Pandas data frame."""
        data_labels = self.data_labels()

        dump_data_loss = False
        for item in self:
            if item.data_loss is not None:
                dump_data_loss = True
                break

        if dump_data_loss:
            columns = ['Timestamp'] + data_labels + ['data_loss']
        else:
            columns = ['Timestamp'] + data_labels

        df = DataFrame(columns=columns)
        for item in self:
            values = [item.data[data_label] for data_label in data_labels]
            if dump_data_loss:
                df = concat([df, DataFrame([[item.dt]+values+[item.data_loss]], columns=columns)])
            else:
                df = concat([df, DataFrame([[item.dt]+values], columns=columns)])
        df = df.set_index('Timestamp')

        return df


    #=========================
    #  Operations
    #=========================

    def slice(self, from_i=None, to_i=None, from_t=None, to_t=None, from_dt=None, to_dt=None):
        """Slice a series. A series of DataPoints or DataSlots is required."""
        from .operations import slice as slice_operation
        return slice_operation(self, from_i=from_i, to_i=to_i, from_t=from_t, to_t=to_t, from_dt=from_dt, to_dt=to_dt)

    def get(self, at_i=None, at_t=None, at_dt=None):
        from .operations import get as get_operation
        return get_operation(self, at_i, at_t, at_dt)

    def view(self, from_i=None, to_i=None):
        """Get a (lazy) view of the series"""
        return TimeSeriesView(series=self, from_i=from_i, to_i=to_i)

    #=========================
    #  Plot-related
    #=========================

    def plot(self, engine='dg', *args, **kwargs):
        """Plot the time series. The default plotting engine is Dygraphs (``engine=\'dg\'``),
           limited support for Matplotplib (``engine=\'mp\'``) is also available.
           For plotting options for Dygraphs, see :func:`~.plots.dygraphs_plot`, while for
           plotting options for Matplotlib, see :func:`~.plots.matplotlib_plot`."""
        if engine=='mp':
            from .plots import matplotlib_plot
            matplotlib_plot(self, *args, **kwargs)
        elif engine=='dg':
            from .plots import dygraphs_plot
            out = dygraphs_plot(self, *args, **kwargs)
            if out: return out
        else:
            raise ValueError('Unknown plotting engine "{}'.format(engine))



#==============================
#  Time Series view
#==============================

class TimeSeriesView(TimeSeries):
    """A time series view."""

    def __init__(self, *items, series=None, from_i=None, to_i=None, **kwargs):
        self.series = series
        self.from_i = from_i
        self.to_i = to_i
        self.len = None
        # Call parent init in case we are using the series view as a normal series
        super().__init__(*items, **kwargs)

    def __getitem__(self, arg):

        if self.series:

            if isinstance(arg, slice):

                if isinstance(arg.start, int):
                    if arg.start>=0:
                        slice_start =  self.from_i + arg.start
                    else:
                        slice_start =  self.to_i - abs(arg.start)
                elif arg.start is None:
                    slice_start = self.from_i
                else:
                    slice_start = arg.start

                if isinstance(arg.stop, int):
                    if arg.stop>=0:
                        slice_stop =  self.from_i + arg.stop
                    else:
                        slice_stop =  self.to_i - abs(arg.stop)
                elif arg.stop is None:
                    slice_stop = self.to_i
                else:
                    slice_stop = arg.stop

                return self.series[slice_start:slice_stop]

            else:
                if isinstance(arg, int):
                    if arg>=0:
                        return self.series[self.from_i + arg]
                    else:
                        return self.series[self.to_i - abs(arg)]
                else:
                    return super().__getitem__(arg)
        else:
            return super().__getitem__(arg)

    def __iter__(self):
        if self.series:
            self.count = 0
            return self
        else:
            return super().__iter__()

    def __next__(self):
        if self.series:
            this_i = self.count + self.from_i

            if this_i >= self.to_i:
                # If reached the end stop
                raise StopIteration

            else:
                # Otherwise return the corresponding item
                self.count += 1
                return self.series[this_i]
        else:
            return super().__next__()

    def __len__(self):
        if self.series:
            return self.to_i-self.from_i
        else:
            return super().__len__()

    def __repr__(self):
        if self.series:
            return 'View from element #{} to element #{} of: {}'.format(self.from_i, self.to_i, self.series)
        else:
            return super().__repr__()

    def __bool__(self):
        if self.series is not None:
            return True if len(self.series) else False
        else:
            return True if len(self) else False

    @property
    def item_type(self):
        if self.series:
            return self.series.item_type
        else:
            return super().item_type

    @property
    def resolution(self):
        if self.series:
            return self.series.resolution
        else:
            return super().resolution

    def data_labels(self):
        if self.series:
            return self.series.data_labels()
        else:
            return super().data_labels()

    def materialize(self):
        """Return a materialized series view"""
        materialized_series = self.__class__()
        for item in self:
            materialized_series.append(deepcopy(item))
        return materialized_series

    def duplicate(self):
        return self.materialize()




# -*- coding: utf-8 -*-
"""Exceptions."""


class ConsistencyException(Exception):
    """Rasied when the internal consistency is broken."""
    pass

class NotFittedError(Exception):
    """Raised when trying to save, apply or evaluate a model that requires fitting first."""
    pass

class AlreadyFittedError(Exception):
    """Raised when trying to fit a model that is already fitted (instead of using the fit update method, if available)."""
    pass

class NonContiguityError(Exception):
    """Raised when the model only supports being applied on data contiguous with the fit data and it is not."""
    pass

class NoDataException(Exception):
    """Raised if a storage get() function finds no  data at all."""
    pass

class FloatConversionError(Exception):
    """Raised to group the various exceptions that can lead to the impossibility of converting a value to a floating point."""
    pass

class NotEnoughDataError(Exception):
    """Raised in context when there is not enough data to perform the required operation"""



